import itertools
import operator
from typing import TYPE_CHECKING, List, Set, Union
from warnings import warn

import rdflib
from rdflib import Literal, URIRef

from pyshacl.consts import (
    RDF,
    SH_alternativePath,
    SH_filterShape,
    SH_intersection,
    SH_inversePath,
    SH_message,
    SH_nodes,
    SH_oneOrMorePath,
    SH_path,
    SH_this,
    SH_union,
    SH_zeroOrMorePath,
    SH_zeroOrOnePath,
)
from pyshacl.errors import ReportableRuntimeError, ShapeLoadError

if TYPE_CHECKING:
    from pyshacl.pytypes import GraphLike, RDFNode
    from pyshacl.shapes_graph import ShapesGraph


def value_nodes_from_path(sg, focus, path_val, target_graph, inverse: bool = False, recursion: int = 0) -> Set:
    # Link: https://www.w3.org/TR/shacl/#property-paths
    if isinstance(path_val, URIRef):
        if inverse:
            return set(target_graph.subjects(path_val, focus))
        else:
            return set(target_graph.objects(focus, path_val))
    elif isinstance(path_val, Literal):
        raise ShapeLoadError(
            "Values of a property path cannot be a Literal.",
            "https://www.w3.org/TR/shacl/#property-paths",
        )
    # At this point, path_val _must_ be a BNode
    # TODO, the path_val BNode must be value of exactly one sh:path subject in the SG.
    if recursion >= 10:
        raise ReportableRuntimeError("Path traversal depth is too much!")
    find_list = set(sg.graph.objects(path_val, RDF.first))
    if len(find_list) > 0:
        first_node = next(iter(find_list))
        rest_nodes = set(sg.graph.objects(path_val, RDF.rest))
        go_deeper = True
        if len(rest_nodes) < 1:
            if recursion == 0:
                raise ReportableRuntimeError("A list of SHACL Paths must contain at least two path items.")
            else:
                go_deeper = False
        rest_node = next(iter(rest_nodes))
        if rest_node == RDF.nil:
            if recursion == 0:
                raise ReportableRuntimeError("A list of SHACL Paths must contain at least two path items.")
            else:
                go_deeper = False
        this_level_nodes = value_nodes_from_path(
            sg, focus, first_node, target_graph, inverse=inverse, recursion=recursion + 1
        )
        if not go_deeper:
            return this_level_nodes
        found_value_nodes = set()
        for tln in iter(this_level_nodes):
            value_nodes = value_nodes_from_path(
                sg, tln, rest_node, target_graph, inverse=inverse, recursion=recursion + 1
            )
            found_value_nodes.update(value_nodes)
        return found_value_nodes

    find_inverse = set(sg.graph.objects(path_val, SH_inversePath))
    if len(find_inverse) > 0:
        inverse_path = next(iter(find_inverse))
        return value_nodes_from_path(
            sg, focus, inverse_path, target_graph, inverse=not inverse, recursion=recursion + 1
        )

    find_alternatives = set(sg.graph.objects(path_val, SH_alternativePath))
    if len(find_alternatives) > 0:
        alternatives_list = next(iter(find_alternatives))
        all_collected = set()
        visited_alternatives = 0
        for a in sg.graph.items(alternatives_list):
            found_nodes = value_nodes_from_path(sg, focus, a, target_graph, inverse=inverse, recursion=recursion + 1)
            visited_alternatives += 1
            all_collected.update(found_nodes)
        if visited_alternatives < 2:
            raise ReportableRuntimeError("List of SHACL alternate paths must have at least two path items.")
        return all_collected

    find_zero_or_more = set(sg.graph.objects(path_val, SH_zeroOrMorePath))
    if len(find_zero_or_more) > 0:
        zm_path = next(iter(find_zero_or_more))
        collection_set = set()
        # Note, the zero-or-more path always includes the current subject too!
        collection_set.add(focus)
        found_nodes = value_nodes_from_path(sg, focus, zm_path, target_graph, inverse=inverse, recursion=recursion + 1)
        search_deeper_nodes = set(iter(found_nodes))
        while len(search_deeper_nodes) > 0:
            current_node = search_deeper_nodes.pop()
            if current_node in collection_set:
                continue
            collection_set.add(current_node)
            found_more_nodes = value_nodes_from_path(
                sg, current_node, zm_path, target_graph, inverse=inverse, recursion=recursion + 1
            )
            search_deeper_nodes.update(found_more_nodes)
        return collection_set

    find_one_or_more = set(sg.graph.objects(path_val, SH_oneOrMorePath))
    if len(find_one_or_more) > 0:
        one_or_more_path = next(iter(find_one_or_more))
        collection_set = set()
        found_nodes = value_nodes_from_path(
            sg, focus, one_or_more_path, target_graph, inverse=inverse, recursion=recursion + 1
        )
        # Note, the one-or-more path should _not_ include the current focus
        search_deeper_nodes = set(iter(found_nodes))
        while len(search_deeper_nodes) > 0:
            current_node = search_deeper_nodes.pop()
            if current_node in collection_set:
                continue
            collection_set.add(current_node)
            found_more_nodes = value_nodes_from_path(
                sg, current_node, one_or_more_path, target_graph, inverse=inverse, recursion=recursion + 1
            )
            search_deeper_nodes.update(found_more_nodes)
        return collection_set

    find_zero_or_one = set(sg.graph.objects(path_val, SH_zeroOrOnePath))
    if len(find_zero_or_one) > 0:
        zero_or_one_path = next(iter(find_zero_or_one))
        collection_set = set()
        # Note, the zero-or-one path always includes the current subject too!
        collection_set.add(focus)
        found_nodes = value_nodes_from_path(
            sg, focus, zero_or_one_path, target_graph, inverse=inverse, recursion=recursion + 1
        )
        collection_set.update(found_nodes)
        return collection_set
    remaining = set(sg.graph.predicate_objects(path_val))
    if len(remaining) > 0:
        raise ShapeLoadError(
            "{} is not a known property for a sh:path. Malformed shape?".format(str(next(iter(remaining))[0])),
            "https://www.w3.org/TR/shacl/#property-paths",
        )

    raise ShapeLoadError(
        "Cannot get any values from sh:path property. Malformed shape?", "https://www.w3.org/TR/shacl/#property-paths"
    )


def nodes_from_node_expression(
    expr, focus_node, data_graph: 'GraphLike', sg: 'ShapesGraph', recurse_depth=0
) -> Union[Set[Union['RDFNode', None]], List[Union['RDFNode', None]]]:
    # https://www.w3.org/TR/shacl-af/#node-expressions
    if expr == SH_this:
        return [focus_node]
    if isinstance(expr, (rdflib.URIRef, rdflib.Literal)):
        return [expr]
    elif isinstance(expr, rdflib.BNode):
        unions = set(sg.objects(expr, SH_union))
        intersections = set(sg.objects(expr, SH_intersection))
        if len(unions) and len(intersections):
            raise ReportableRuntimeError("Cannot have sh:intersection and sh:union on the same bnode.")
        if recurse_depth > 8 and (len(unions) or len(intersections)):
            warn(Warning("sh:union, sh:intersection, or sh:function args depth too deep. Won't capture all of it!"))
            return []
        if len(unions):
            union_list = next(iter(unions))
            parts = list(sg.graph.items(union_list))
            all_nodes: Set[Union['RDFNode', None]] = set()
            for p in parts:
                new_parts = nodes_from_node_expression(p, focus_node, data_graph, sg, recurse_depth=recurse_depth + 1)
                all_nodes = all_nodes.union(new_parts)
            return all_nodes
        if len(intersections):
            inter_list = next(iter(intersections))
            parts = list(data_graph.items(inter_list))
            inter_nodes: Set[Union['RDFNode', None]] = set()
            new = True
            for p in parts:
                new_parts = nodes_from_node_expression(p, focus_node, data_graph, sg, recurse_depth=recurse_depth + 1)
                if new is True:
                    inter_nodes = set(iter(new_parts))
                    new = False
                else:
                    inter_nodes = inter_nodes.intersection(new_parts)
            return inter_nodes
        path_nodes = set(sg.objects(expr, SH_path))
        if len(path_nodes) > 0:
            path_results: Set[Union['RDFNode', None]] = set()
            for p in path_nodes:
                vals = value_nodes_from_path(sg, focus_node, p, data_graph)
                path_results.update(vals)
            return path_results
        filter_shapes = set(sg.objects(expr, SH_filterShape))
        nodes_nodes = set(sg.objects(expr, SH_nodes))
        if len(filter_shapes) > 0:  # pragma: no cover
            # Note: There's no tests for this whole filterShapes feature!
            if len(nodes_nodes) > 1:
                warn(Warning("More than one sh:nodes found. Using the first one."))
            elif len(nodes_nodes) < 1:
                raise ReportableRuntimeError("The Node FilterShape {} does not have sh:nodes.".format(expr))
            filter_shape = next(iter(filter_shapes))
            filter_shape = sg.lookup_shape_from_node(filter_shape)
            nodes_expr = next(iter(nodes_nodes))
            to_filter = nodes_from_node_expression(
                nodes_expr, focus_node, data_graph, sg, recurse_depth=recurse_depth + 1
            )
            passes = set()
            for n in to_filter:
                conforms, reports = filter_shape.validate(data_graph, n)
                if conforms:
                    passes.add(n)
            return passes
        # Got to here, the only other possibility is this is a FunctionExpression.
        remain_pairs = set(sg.predicate_objects(expr))
        fn_pairs = set()
        for rk, rv in remain_pairs:
            if rk == SH_message:
                # expressions can have a message, it doesn't affect the function
                continue
            elif isinstance(rv, rdflib.Literal):
                # pair is not list-valued, it can't be a function
                continue
            else:
                has_first = list(sg.objects(rv, RDF.first))
                if len(has_first) > 0:
                    fn_pairs.add((rk, rv))
        if len(fn_pairs) > 1:
            warn(Warning("More than one FunctionExpression found. Using the first one."))
        fnexpr, fnargslist = next(iter(fn_pairs))
        # find the function!
        try:
            function, optionals = sg.get_shacl_function(fnexpr)
        except KeyError:
            raise ReportableRuntimeError(
                "The SHACLFunction {} was not defined in this SHACL Shapes file.".format(fnexpr)
            )
        argslist_parts = list(sg.graph.items(fnargslist))
        args_sets: List[Union[List[Union['RDFNode', None]], Set[Union['RDFNode', None]]]] = [
            nodes_from_node_expression(p, focus_node, data_graph, sg, recurse_depth=recurse_depth + 1)
            for p in argslist_parts
        ]
        num_args_sets = len(args_sets)
        num_expected_args = len(optionals)
        if num_args_sets > num_expected_args:
            raise ReportableRuntimeError("Too many arguments given for {}".format(fnexpr))
        if num_args_sets < num_expected_args:
            # not enough, but some might be optional
            num_diff = num_expected_args - num_args_sets
            args_sets_slice = args_sets[0 - num_diff :]
            all_empty = itertools.accumulate(
                {True if len(a) < 1 else False for a in args_sets_slice}, func=operator.or_
            )
            if all_empty:
                args_sets = args_sets[:num_expected_args]
            else:
                raise ReportableRuntimeError(
                    "Too few arguments given for {}, with non-optional arguments.".format(fnexpr)
                )
        add_nones = set()
        for i, a in enumerate(args_sets):
            if len(a) < 1:
                if optionals[i] is False:
                    warn(Warning("Got an empty set of nodes for a non-optional argument in {}.".format(fnexpr)))
                    return []
                else:
                    add_nones.add(i)
        for i in add_nones:
            args_sets[i] = [None]
        args_permutations = list(itertools.product(*args_sets))
        responses = set()
        for permus in args_permutations:
            result = function(data_graph, *permus)
            responses.add(result)
        return responses
    else:
        raise NotImplementedError("Unsupported expression {}".format(expr))



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

import argparse
import sys
from io import BufferedReader
from typing import Union, cast

from pyshacl import __version__, shacl_rules
from pyshacl.cli import ShowVersion
from pyshacl.errors import (
    ConstraintLoadError,
    ReportableRuntimeError,
    RuleLoadError,
    ShapeLoadError,
    ValidationFailure,
)

parser = argparse.ArgumentParser(
    description='PySHACL {} SHACL Rules Expander command line tool.'.format(str(__version__))
)
parser.add_argument(
    'data',
    metavar='DataGraph',
    help='The file or endpoint containing the Target Data Graph.',
    default=None,
    nargs='?',
)
parser.add_argument(
    '-s',
    '--shapes',
    '--shacl',
    dest='shacl',
    action='store',
    nargs='?',
    help='A file containing the SHACL Shapes Graph.',
)
parser.add_argument(
    '-e',
    '--ont-graph',
    dest='ont',
    action='store',
    nargs='?',
    help='A file path or URL to a document containing extra ontological information. '
    'RDFS and OWL definitions from this are used to inoculate the DataGraph.',
)
parser.add_argument(
    '-i',
    '--inference',
    dest='inference',
    action='store',
    default='none',
    choices=('none', 'rdfs', 'owlrl', 'both'),
    help='Choose a type of inferencing to run against the Data Graph before validating.',
)
parser.add_argument(
    '-im',
    '--imports',
    dest='imports',
    action='store_true',
    default=False,
    help='Allow import of sub-graphs defined in statements with owl:imports.',
)
parser.add_argument(
    '-j',
    '--js',
    dest='js',
    action='store_true',
    default=False,
    help='Enable features from the SHACL-JS Specification.',
)
parser.add_argument(
    '-it',
    '--iterate-rules',
    dest='iterate_rules',
    action='store_true',
    default=False,
    help="Run Shape's SHACL Rules iteratively until the data_graph reaches a steady state.",
)
parser.add_argument(
    '-d',
    '--debug',
    dest='debug',
    action='store_true',
    default=False,
    help='Output additional verbose runtime messages.',
)
parser.add_argument(
    '--focus',
    dest='focus',
    action='store',
    help='Optional IRIs of focus nodes from the DataGraph, the shapes will validate only these node. Comma-separated list.',
    nargs="?",
    default=None,
)
parser.add_argument(
    '--shape',
    dest='shape',
    action='store',
    help='Optional IRIs of a NodeShape or PropertyShape from the SHACL ShapesGraph, only these shapes will be used to validate the DataGraph. Comma-separated list.',
    nargs="?",
    default=None,
)
parser.add_argument(
    '-f',
    '--format',
    dest='format',
    action='store',
    help='Choose an output format. Default is "trig" for Datasets and "turtle" for Graphs.',
    default='auto',
    choices=('auto', 'turtle', 'xml', 'trig', 'json-ld', 'nt', 'n3', 'nquads'),
)
parser.add_argument(
    '-df',
    '--data-file-format',
    dest='data_file_format',
    action='store',
    help='Explicitly state the RDF File format of the input DataGraph file. Default="auto".',
    default='auto',
    choices=('auto', 'turtle', 'xml', 'trig', 'json-ld', 'nt', 'n3', 'nquads'),
)
parser.add_argument(
    '-sf',
    '--shacl-file-format',
    dest='shacl_file_format',
    action='store',
    help='Explicitly state the RDF File format of the input SHACL file. Default="auto".',
    default='auto',
    choices=('auto', 'turtle', 'xml', 'trig', 'json-ld', 'nt', 'n3', 'nquads'),
)
parser.add_argument(
    '-ef',
    '--ont-file-format',
    dest='ont_file_format',
    action='store',
    help='Explicitly state the RDF File format of the extra ontology file. Default="auto".',
    default='auto',
    choices=('auto', 'turtle', 'xml', 'trig', 'json-ld', 'nt', 'n3', 'nquads'),
)
parser.add_argument('-V', '--version', action=ShowVersion, help='Show PySHACL version and exit.')
parser.add_argument(
    '-o',
    '--output',
    dest='output',
    nargs='?',
    type=argparse.FileType('w'),
    help='Send output to a file (defaults to stdout).',
    default=sys.stdout,
)
# parser.add_argument('-h', '--help', action="help", help='Show this help text.')


def main(prog: Union[str, None] = None) -> None:
    if prog is not None and len(prog) > 0:
        parser.prog = prog

    args = parser.parse_args()
    if not args.data:
        # No datafile give, and not starting in server mode.
        sys.stderr.write('Input Error. No DataGraph file or endpoint supplied.\n')
        parser.print_usage(sys.stderr)
        sys.exit(1)
    runner_kwargs = {
        'debug': args.debug,
        'serialize_expanded_graph': True,
    }
    data_file = None
    data_graph: Union[BufferedReader, str]

    try:
        data_file = open(args.data, 'rb')
    except FileNotFoundError:
        sys.stderr.write('Input Error. DataGraph file not found.\n')
        sys.exit(1)
    except PermissionError:
        sys.stderr.write('Input Error. DataGraph file not readable.\n')
        sys.exit(1)
    else:
        # NOTE: This cast is not necessary in Python >= 3.10.
        data_graph = cast(BufferedReader, data_file)
    if args.shacl is not None:
        runner_kwargs['shacl_graph'] = args.shacl
    if args.ont is not None:
        runner_kwargs['ont_graph'] = args.ont
    if args.inference != 'none':
        runner_kwargs['inference'] = args.inference
    if args.imports:
        runner_kwargs['do_owl_imports'] = True
    if args.js:
        runner_kwargs['js'] = True
    if args.focus:
        runner_kwargs['focus_nodes'] = [_f.strip() for _f in args.focus.split(',')]
    if args.shape:
        runner_kwargs['use_shapes'] = [_s.strip() for _s in args.shape.split(',')]
    if args.iterate_rules:
        runner_kwargs['iterate_rules'] = True
    if args.shacl_file_format:
        _f: str = args.shacl_file_format
        if _f != "auto":
            runner_kwargs['shacl_graph_format'] = _f
    if args.ont_file_format:
        _f = args.ont_file_format
        if _f != "auto":
            runner_kwargs['ont_graph_format'] = _f
    if args.data_file_format:
        _f = args.data_file_format
        if _f != "auto":
            runner_kwargs['data_graph_format'] = _f
    if args.format != "auto":
        runner_kwargs['serialize_expanded_graph_format'] = args.format
    exit_code: Union[int, None] = None
    try:
        output_txt = shacl_rules(data_graph, **runner_kwargs)
        if isinstance(output_txt, BaseException):
            raise output_txt
    except ValidationFailure as vf:
        args.output.write("Rules Runner generated a Validation Failure result:\n")
        args.output.write(str(vf.message))
        args.output.write("\n")
        exit_code = 1
    except ShapeLoadError as sle:
        sys.stderr.write("Rules Runner encountered a Shape Load Error:\n")
        sys.stderr.write(str(sle))
        exit_code = 2
    except ConstraintLoadError as cle:
        sys.stderr.write("Rules Runner encountered a Constraint Load Error:\n")
        sys.stderr.write(str(cle))
        exit_code = 2
    except RuleLoadError as rle:
        sys.stderr.write("Rules Runner encountered a Rule Load Error:\n")
        sys.stderr.write(str(rle))
        exit_code = 2
    except ReportableRuntimeError as rre:
        sys.stderr.write("Rules Runner encountered a Runtime Error:\n")
        sys.stderr.write(str(rre.message))
        sys.stderr.write("\nIf you believe this is a bug in pyshacl, open an Issue on the pyshacl github page.\n")
        exit_code = 2
    except NotImplementedError as nie:
        sys.stderr.write("Rules Runner feature is not implemented:\n")
        if len(nie.args) > 0:
            sys.stderr.write(str(nie.args[0]))
        else:
            sys.stderr.write("No message provided.")
        sys.stderr.write("\nIf your use-case requires this feature, open an Issue on the pyshacl github page.\n")
        exit_code = 3
    except RuntimeError as re:
        import traceback

        traceback.print_tb(re.__traceback__)
        sys.stderr.write(
            "\n\nRules Runner encountered a Runtime Error. Please report this to the PySHACL issue tracker.\n"
        )
        exit_code = 2
    finally:
        if data_file is not None:
            try:
                data_file.close()
            except Exception as e:
                sys.stderr.write("Error closing data file:\n")
                sys.stderr.write(str(e))
        if exit_code is not None:
            sys.exit(exit_code)

        if isinstance(output_txt, bytes):
            output_unicode = output_txt.decode('utf-8')
        else:
            output_unicode = output_txt
        args.output.write(output_unicode)
    args.output.close()
    sys.exit(0)


if __name__ == "__main__":
    main()



# -*- coding: utf-8 -*-
import logging
import warnings
from typing import TYPE_CHECKING, Dict, List, Optional, Sequence, Union

import rdflib

from .constraints.constraint_component import CustomConstraintComponentFactory
from .constraints.core.logical_constraints import SH_and, SH_not, SH_or, SH_xone
from .constraints.core.shape_based_constraints import SH_qualifiedValueShape
from .consts import (
    SH,
    OWL_Class,
    OWL_DatatypeProperty,
    RDF_Property,
    RDF_type,
    RDFS_Class,
    RDFS_subClassOf,
    SH_ConstraintComponent,
    SH_node,
    SH_NodeShape,
    SH_path,
    SH_property,
    SH_PropertyShape,
    SH_targetClass,
    SH_targetNode,
    SH_targetObjectsOf,
    SH_targetSubjectsOf,
)
from .errors import ShapeLoadError
from .shape import Shape

if TYPE_CHECKING:
    from .pytypes import RDFNode


class ShapesGraph(object):
    system_triples = [(OWL_Class, RDFS_subClassOf, RDFS_Class), (OWL_DatatypeProperty, RDFS_subClassOf, RDF_Property)]

    def __init__(self, graph, debug: Optional[Union[bool, int]] = False, logger: Optional[logging.Logger] = None):
        """
        ShapesGraph
        :param graph:
        :type graph: rdflib.Graph
        :param debug:
        :type debug: bool|int|None
        :param logger:
        :type logger: logging.Logger|None
        """
        assert isinstance(graph, (rdflib.Dataset, rdflib.ConjunctiveGraph, rdflib.Graph))
        self.graph = graph
        if isinstance(self.graph, rdflib.Dataset):
            self.graph.default_union = True
        if logger is None:
            logger = logging.getLogger(__name__)
        self.logger = logger
        self.debug = debug
        self._node_shape_cache: Dict['RDFNode', Shape] = {}
        self._shapes = None
        self._custom_constraints = None
        self._shacl_functions: Dict[str, tuple] = {}
        self._shacl_target_types: Dict[str, 'RDFNode'] = {}
        self._use_js = False
        self._add_system_triples()

    def enable_js(self):
        self._use_js = True

    @property
    def js_enabled(self):
        return bool(self._use_js)

    def _add_system_triples(self):
        if isinstance(self.graph, (rdflib.Dataset, rdflib.ConjunctiveGraph)):
            g = self.graph.default_context
        else:
            g = self.graph
        for t in self.system_triples:
            g.add(t)

    def subjects(self, p, o):
        return self.graph.subjects(p, o)

    def predicates(self, s, o):
        return self.graph.predicates(s, o)

    def objects(self, s, p):
        return self.graph.objects(s, p)

    def subject_objects(self, p):
        return self.graph.subject_objects(p)

    def subject_predicates(self, o):
        return self.graph.subject_predicates(o)

    def predicate_objects(self, s):
        return self.graph.predicate_objects(s)

    @property
    def custom_constraints(self):
        if self._custom_constraints is None:
            self._custom_constraints = self._find_custom_constraints()
        return self._custom_constraints

    def _find_custom_constraints(self):
        g = self.graph
        constraint_component_set = set(g.subjects(RDF_type, SH_ConstraintComponent))
        constraint_subclasses = set(g.subjects(RDFS_subClassOf, SH_ConstraintComponent))
        for sc in iter(constraint_subclasses):
            subclass_components = set(g.subjects(RDF_type, sc))
            constraint_component_set.update(subclass_components)
        components = set()
        for c in iter(constraint_component_set):
            if c.startswith(SH):
                # ignore all constraint components from shacl.ttl, these are all hardcoded into PySHACL
                continue
            components.add(CustomConstraintComponentFactory(self, c))
        return components

    def add_shacl_function(self, uri: Union[str, 'RDFNode'], function, optionals):
        uri_str = str(uri)
        if uri_str in self._shacl_functions:
            warnings.warn(Warning("SHACLFunction {} is already registered.".format(uri)))
        else:
            self._shacl_functions[uri_str] = (function, optionals)

    def get_shacl_function(self, uri: Union[str, 'RDFNode']):
        uri_str = str(uri)
        try:
            f = self._shacl_functions[uri_str]
        except LookupError:
            raise KeyError("SHACLFunction {} not found.".format(uri))
        return f

    def remove_shacl_function(self, uri, function):
        uri = str(uri)
        try:
            f, _ = self.get_shacl_function(uri)
            if f != function:
                warnings.warn(
                    Warning("Cannot remove a different function than what was registered for {}.".format(uri))
                )
                return
        except KeyError:
            return  # Not registered
        del self._shacl_functions[uri]

    def add_shacl_target_type(self, uri: Union[str, 'RDFNode'], tt):
        uri_str = str(uri)
        if uri_str in self._shacl_target_types:
            warnings.warn(Warning("SHACL TargetType {} is already registered.".format(uri)))
        else:
            self._shacl_target_types[uri_str] = tt

    def get_shacl_target_type(self, uri: Union[str, 'RDFNode']):
        uri_str = str(uri)
        try:
            tt = self._shacl_target_types[uri_str]
        except LookupError:
            raise KeyError("SHACL TargetType {} not found.".format(uri))
        return tt

    @property
    def shapes(self):
        """

        :returns: List[Shape]
        :rtype: List[pyshacl.shape.Shape]
        """
        if len(self._node_shape_cache) < 1:
            self._build_node_shape_cache()
        return self._node_shape_cache.values()

    def shapes_from_uris(self, shapes_uris: List[rdflib.URIRef]):
        """
        :param shapes_uris:
        :type shapes_uris: List[rdflib.URIRef]
        :returns: List[Shape]
        :rtype: List[Shape]
        """
        if len(self._node_shape_cache) < 1:
            self._build_node_shape_cache_from_list(shapes_uris)
        return [self._node_shape_cache[s] for s in shapes_uris]

    def lookup_shape_from_node(self, node) -> Shape:
        # This will throw a KeyError if it is not found. This is intentionally not caught here.
        return self._node_shape_cache[node]

    """
    A shape is an IRI or blank node s that fulfills at least one of the following conditions in the shapes graph:

        s is a SHACL instance of sh:NodeShape or sh:PropertyShape.
        s is subject of a triple that has sh:targetClass, sh:targetNode, sh:targetObjectsOf or sh:targetSubjectsOf as predicate.
        s is subject of a triple that has a parameter as predicate.
        s is a value of a shape-expecting, non-list-taking parameter such as sh:node, or a member of a SHACL list that is a value of a shape-expecting and list-taking parameter such as sh:or.
    """

    def _build_node_shape_cache(self):
        """
        :returns: None
        :rtype: NoneType
        """
        g = self.graph
        defined_node_shapes = set(g.subjects(RDF_type, SH_NodeShape))
        if self.debug:
            self.logger.debug(f"Found {len(defined_node_shapes)} SHACL Shapes defined with type sh:NodeShape.")
        for s in defined_node_shapes:
            path_vals = list(g.objects(s, SH_path))
            if len(path_vals) > 0:
                # TODO:coverage: we don't have any tests for invalid shapes
                raise ShapeLoadError(
                    "A shape defined as a NodeShape cannot be the subject of a 'sh:path' predicate.",
                    "https://www.w3.org/TR/shacl/#node-shapes",
                )
        defined_prop_shapes = set(g.subjects(RDF_type, SH_PropertyShape))
        if self.debug:
            self.logger.debug(f"Found {len(defined_prop_shapes)} SHACL Shapes defined with type sh:PropertyShape.")
        found_prop_shapes_paths = dict()
        for s in defined_prop_shapes:
            if s in defined_node_shapes:
                # TODO:coverage: we don't have any tests for invalid shapes
                raise ShapeLoadError(
                    "A shape defined as a NodeShape cannot also be defined as a PropertyShape.",
                    "https://www.w3.org/TR/shacl/#node-shapes",
                )
            path_vals = list(g.objects(s, SH_path))
            if len(path_vals) < 1:
                # TODO:coverage: we don't have any tests for invalid shapes
                raise ShapeLoadError(
                    "A shape defined as a PropertyShape must include one `sh:path` property.",
                    "https://www.w3.org/TR/shacl/#property-shapes",
                )
            elif len(path_vals) > 1:
                # TODO:coverage: we don't have any tests for invalid shapes
                raise ShapeLoadError(
                    "A shape defined as a PropertyShape cannot have more than one 'sh:path' property.",
                    "https://www.w3.org/TR/shacl/#property-shapes",
                )
            found_prop_shapes_paths[s] = path_vals[0]
        if self.debug:
            self.logger.debug(f"Found {len(found_prop_shapes_paths)} property paths to follow.")
        has_target_class = {s for s, o in g.subject_objects(SH_targetClass)}
        has_target_node = {s for s, o in g.subject_objects(SH_targetNode)}
        has_target_objects_of = {s for s, o in g.subject_objects(SH_targetObjectsOf)}
        has_target_subjects_of = {s for s, o in g.subject_objects(SH_targetSubjectsOf)}
        subject_shapes = set(has_target_class).union(
            set(has_target_node).union(set(has_target_objects_of).union(set(has_target_subjects_of)))
        )
        # implicit shapes: their subjects must be shapes
        subject_of_property = {s for s, o in g.subject_objects(SH_property)}
        subject_of_node = {s for s, o in g.subject_objects(SH_node)}
        subject_shapes = subject_shapes.union(set(subject_of_property).union(set(subject_of_node)))
        if self.debug:
            self.logger.debug(f"Found {len(subject_shapes)} implied SHACL Shapes based on their properties.")

        # shape-expecting properties, their values must be shapes.
        value_of_property = {o for s, o in g.subject_objects(SH_property)}
        value_of_node = {o for s, o in g.subject_objects(SH_node)}
        value_of_not = {o for s, o in g.subject_objects(SH_not)}
        value_of_qvs = {o for s, o in g.subject_objects(SH_qualifiedValueShape)}
        value_of_shape_expecting = set(value_of_property).union(
            set(value_of_node).union(set(value_of_not).union(set(value_of_qvs)))
        )

        value_of_and = {o for s, o in g.subject_objects(SH_and)}
        value_of_or = {o for s, o in g.subject_objects(SH_or)}
        value_of_xone = {o for s, o in g.subject_objects(SH_xone)}
        value_of_s_list_expecting = set(value_of_and).union(set(value_of_or).union(set(value_of_xone)))

        for lst in value_of_s_list_expecting:
            list_contents = set(g.items(lst))
            if len(list_contents) < 1:
                # TODO:coverage: we don't have any tests for invalid shape lists
                raise ShapeLoadError(
                    "A Shape-Expecting & List-Expecting predicate should get a well-formed RDF list with 1 or more members.",
                    "https://www.w3.org/TR/shacl/#shapes-recursion",
                )
            for s in list_contents:
                value_of_shape_expecting.add(s)

        if self.debug:
            self.logger.debug(
                f"Found {len(value_of_shape_expecting)} implied SHACL Shapes used as values in shape-expecting constraints."
            )
        found_node_shapes = set()
        found_prop_shapes = set()
        for s in subject_shapes:
            if s in defined_node_shapes or s in defined_prop_shapes:
                continue
            path_vals = list(g.objects(s, SH_path))
            if len(path_vals) < 1:
                found_node_shapes.add(s)
            elif len(path_vals) > 1:
                # TODO:coverage: we don't have any tests for invalid implicit shapes
                raise ShapeLoadError(
                    "An implicit PropertyShape cannot have more than one 'sh:path' predicate.",
                    "https://www.w3.org/TR/shacl/#property-shapes",
                )
            else:
                # TODO:coverage: we don't have this case where an implicit shape is a property shape.
                found_prop_shapes.add(s)
                found_prop_shapes_paths[s] = path_vals[0]
        for s in value_of_shape_expecting:
            if (
                s in defined_node_shapes
                or s in defined_prop_shapes
                or s in found_prop_shapes
                or s in found_node_shapes
            ):
                continue
            path_vals = list(g.objects(s, SH_path))
            if len(path_vals) < 1:
                found_node_shapes.add(s)
            elif len(path_vals) > 1:
                # TODO:coverage: we don't have any tests for invalid implicit shapes
                raise ShapeLoadError(
                    "An implicit PropertyShape cannot have more than one 'sh:path' predicate.",
                    "https://www.w3.org/TR/shacl/#property-shapes",
                )
            else:
                found_prop_shapes.add(s)
                found_prop_shapes_paths[s] = path_vals[0]
        node_shape_count = 0
        property_shape_count = 0
        for node_shape in defined_node_shapes.union(found_node_shapes):
            if node_shape in self._node_shape_cache:
                # TODO:coverage: we don't have any tests where a shape is loaded twice
                raise ShapeLoadError("That shape has already been loaded!", "None")
            s = Shape(self, node_shape, p=False, logger=self.logger)
            self._node_shape_cache[node_shape] = s
            node_shape_count += 1
        for prop_shape in defined_prop_shapes.union(found_prop_shapes):
            if prop_shape in self._node_shape_cache:
                # TODO:coverage: we don't have any tests where a shape is loaded twice
                raise ShapeLoadError("That shape has already been loaded!", "None")
            prop_shape_path = found_prop_shapes_paths[prop_shape]
            s = Shape(self, prop_shape, p=True, path=prop_shape_path, logger=self.logger)
            self._node_shape_cache[prop_shape] = s
            property_shape_count += 1
        if self.debug:
            self.logger.debug(
                f"Cached {node_shape_count} unique NodeShapes and {property_shape_count} unique PropertyShapes."
            )

    def _build_node_shape_cache_from_list(self, shapes_list: List[rdflib.URIRef]):
        """
        :returns: None
        :rtype: NoneType
        """
        g = self.graph
        gathered_node_shapes = set()
        gathered_prop_shapes = set()
        found_prop_shapes_paths: Dict[Union[rdflib.URIRef, rdflib.BNode], Union[rdflib.URIRef, rdflib.BNode]] = dict()

        def _gather_shapes(shapes_nodes: Sequence[Union[rdflib.URIRef, rdflib.BNode]], recurse_depth: int = 0):
            nonlocal gathered_node_shapes, gathered_prop_shapes, found_prop_shapes_paths
            if recurse_depth > 10:
                raise ShapeLoadError(
                    "Specified shape has too many levels of attached bnodes.",
                    "https://www.w3.org/TR/shacl/#shapes-graph",
                )
            shape_expecting_preds = (SH_and, SH_not, SH_or, SH_xone, SH_property, SH_node, SH_qualifiedValueShape)
            for s in shapes_nodes:
                all_po = list(g.predicate_objects(s))
                if len(all_po) < 1:
                    if recurse_depth < 1:
                        raise ShapeLoadError(
                            "Shape listed in use_shapes does not exist in the SHACL ShapesGraph.",
                            "https://www.w3.org/TR/shacl/#shapes-graph",
                        )
                    else:
                        return
                has_class = any(RDF_type == _p for _p, _o in all_po)
                has_shape_expecting_p: Dict[rdflib.URIRef, bool] = {}
                for _p in shape_expecting_preds:
                    if any(_p == _p2 for _p2, _o in all_po):
                        has_shape_expecting_p[_p] = True

                knows_class = False
                if has_class:
                    all_classes = list(g.objects(s, RDF_type))
                    for c in all_classes:
                        if c == SH_PropertyShape:
                            knows_class = True
                            gathered_prop_shapes.add(s)
                            break
                        elif c == SH_NodeShape:
                            knows_class = True
                            gathered_node_shapes.add(s)
                            break
                    else:
                        knows_class = False
                if not knows_class:
                    for po_p, po_o in all_po:
                        if po_p == SH_path:
                            if not isinstance(po_o, (rdflib.BNode, rdflib.URIRef)):
                                raise ShapeLoadError(
                                    "Found a path property with a value that is not a URIRef or BNode.",
                                    "https://www.w3.org/TR/shacl/#property-paths",
                                )
                            has_path = True
                            found_prop_shapes_paths[s] = po_o
                            break
                    else:
                        has_path = False
                    if has_path:
                        gathered_prop_shapes.add(s)
                    else:
                        is_property_of = len(list(g.subjects(SH_property, s))) > 0
                        if is_property_of:
                            gathered_prop_shapes.add(s)
                        else:
                            gathered_node_shapes.add(s)
                _found_child_bnodes: List[rdflib.BNode] = []
                if has_shape_expecting_p:
                    for _p in has_shape_expecting_p.keys():
                        property_entries = list(g.objects(s, _p))
                        for p_e in property_entries:
                            if _p in (SH_or, SH_xone, SH_and):
                                # These are list-expecting variants of shape-expecting constraints.
                                for item in g.items(p_e):
                                    if isinstance(item, rdflib.BNode):
                                        _found_child_bnodes.append(item)
                            elif isinstance(p_e, rdflib.BNode):
                                _found_child_bnodes.append(p_e)
                if len(_found_child_bnodes) > 0:
                    _gather_shapes(_found_child_bnodes, recurse_depth=recurse_depth + 1)

        _gather_shapes(shapes_list)

        for s in gathered_node_shapes:
            path_vals = list(g.objects(s, SH_path))
            if len(path_vals) > 0:
                # TODO:coverage: we don't have any tests for invalid shapes
                raise ShapeLoadError(
                    "A shape defined as a NodeShape cannot be the subject of a 'sh:path' predicate.",
                    "https://www.w3.org/TR/shacl/#node-shapes",
                )

        for s in gathered_prop_shapes:
            if s in gathered_node_shapes:
                # TODO:coverage: we don't have any tests for invalid shapes
                raise ShapeLoadError(
                    "A shape defined as a NodeShape cannot also be defined as a PropertyShape.",
                    "https://www.w3.org/TR/shacl/#node-shapes",
                )
            if s not in found_prop_shapes_paths:
                path_vals = list(g.objects(s, SH_path))
                if len(path_vals) < 1:
                    # TODO:coverage: we don't have any tests for invalid shapes
                    raise ShapeLoadError(
                        "A shape defined as a PropertyShape must include one `sh:path` property.",
                        "https://www.w3.org/TR/shacl/#property-shapes",
                    )
                elif len(path_vals) > 1:
                    # TODO:coverage: we don't have any tests for invalid shapes
                    raise ShapeLoadError(
                        "A shape defined as a PropertyShape cannot have more than one 'sh:path' property.",
                        "https://www.w3.org/TR/shacl/#property-shapes",
                    )
                else:
                    _p = path_vals[0]
                    if not isinstance(_p, (rdflib.BNode, rdflib.URIRef)):
                        raise ShapeLoadError(
                            "Found a path property with a value that is not a URIRef or BNode.",
                            "https://www.w3.org/TR/shacl/#property-paths",
                        )
                    found_prop_shapes_paths[s] = _p

        node_shape_count = 0
        property_shape_count = 0
        for node_shape in gathered_node_shapes:
            if node_shape in self._node_shape_cache:
                # TODO:coverage: we don't have any tests where a shape is loaded twice
                raise ShapeLoadError("That shape has already been loaded!", "None")
            _s = Shape(self, node_shape, p=False, logger=self.logger)
            self._node_shape_cache[node_shape] = _s
            node_shape_count += 1
        for prop_shape in gathered_prop_shapes:
            if prop_shape in self._node_shape_cache:
                # TODO:coverage: we don't have any tests where a shape is loaded twice
                raise ShapeLoadError("That shape has already been loaded!", "None")
            prop_shape_path = found_prop_shapes_paths[prop_shape]
            _s = Shape(self, prop_shape, p=True, path=prop_shape_path, logger=self.logger)
            self._node_shape_cache[prop_shape] = _s
            property_shape_count += 1

        if self.debug:
            self.logger.debug(
                f"Cached {node_shape_count} unique NodeShapes and {property_shape_count} unique PropertyShapes."
            )



# -*- coding: utf-8 -*-
#

# This software was developed at the National Institute of Standards
# and Technology by employees of the Federal Government in the course
# of their official duties. Pursuant to title 17 Section 105 of the
# United States Code this software is not subject to copyright
# protection and is in the public domain. NIST assumes no
# responsibility whatsoever for its use by other parties, and makes
# no guarantees, expressed or implied, about its quality,
# reliability, or any other characteristic.
#
# We would appreciate acknowledgement if the software is used.

"""
https://github.com/RDFLib/pySHACL/issues/116
"""

import rdflib

from pyshacl import validate


shacl_file_base = '''
@prefix rdf: <http://www.w3.org/1999/02/22-rdf-syntax-ns#> .
@prefix rdfs: <http://www.w3.org/2000/01/rdf-schema#> .
@prefix sh: <http://www.w3.org/ns/shacl#> .
@prefix xml: <http://www.w3.org/XML/1998/namespace> .
@prefix xsd: <http://www.w3.org/2001/XMLSchema#> .
@prefix owl: <http://www.w3.org/2002/07/owl#> .
@prefix ex: <http://example.org/issue/116#> .

<http://example.org/issue/116>
  rdf:type owl:Ontology ;
  sh:declare [
    sh:prefix "ex" ;
    sh:namespace "http://example.org/issue/116#"^^xsd:anyURI ;
    ] .

ex:ThingWithAStringProperty
	a owl:Class ;
	a sh:NodeShape ;
	sh:property [
		sh:datatype %s ;
		sh:path ex:someString ;
	] .
'''

shacl_file_with_plain_literal = shacl_file_base % "rdf:langString"
shacl_file_with_string = shacl_file_base % "xsd:string"

data_file_base = """
@prefix rdf: <http://www.w3.org/1999/02/22-rdf-syntax-ns#> .
@prefix rdfs: <http://www.w3.org/2000/01/rdf-schema#> .
@prefix xml: <http://www.w3.org/XML/1998/namespace> .
@prefix xsd: <http://www.w3.org/2001/XMLSchema#> .
@prefix ex: <http://example.org/issue/116#> .
@prefix kb: <http://example.org/kb/> .

kb:someIndividual
	a ex:ThingWithAStringProperty ;
"""

data_file_plain_literal = data_file_base + """	ex:someString "A string with a language"@en ."""
data_file_string = data_file_base + """	ex:someString "A string without a language" ."""


def _test_116_template(shacl_file_text: str, data_file_text: str, should_conform: bool) -> None:
    data = rdflib.Graph()
    data.parse(data=data_file_text, format="turtle")
    res = validate(
        data,
        shacl_graph=shacl_file_text,
        data_graph_format='turtle',
        shacl_graph_format='turtle',
        debug=True,
    )
    conforms, graph, string = res
    assert should_conform == conforms


def test_116_1():
    _test_116_template(shacl_file_with_string, data_file_plain_literal, False)


def test_116_2():
    _test_116_template(shacl_file_with_string, data_file_string, True)


def test_116_3():
    _test_116_template(shacl_file_with_plain_literal, data_file_plain_literal, True)


def test_116_4():
    _test_116_template(shacl_file_with_plain_literal, data_file_string, False)


if __name__ == "__main__":
    test_116_1()
    test_116_2()
    test_116_3()
    test_116_4()



# -*- coding: utf-8 -*-
#
"""
https://github.com/RDFLib/pySHACL/issues/87
"""
from pyshacl import validate

mixed_file_text = """
@prefix sh: <http://www.w3.org/ns/shacl#> .
@prefix rdf: <http://www.w3.org/1999/02/22-rdf-syntax-ns#> .
@prefix owl: <http://www.w3.org/2002/07/owl#> .
@prefix rdfs: <http://www.w3.org/2000/01/rdf-schema#> .
@prefix : <urn:ex#> .

:TargetClass a owl:Class .

:Class0 a owl:Class .

:Class1 a owl:Class ;
    rdfs:subClassOf :Class0 .

:Class2 a owl:Class ;
    rdfs:subClassOf :Class1 .

:Class3 a owl:Class ;
    rdfs:subClassOf :Class2 .

:prop  a   owl:ObjectProperty .

:shape a sh:NodeShape ;
    sh:targetClass :TargetClass ;
    sh:property :haspointshape .

:haspointshape a sh:PropertyShape ;
    sh:path :prop ;
    sh:class :Class0 .

:s0 a :Class0 .
:vav0 a :TargetClass ;
    :prop :s0 .

:s1 a :Class1 .
:vav1 a :TargetClass ;
    :prop :s1 .

:s2 a :Class2 .
:vav2 a :TargetClass ;
    :prop :s2 .

:s3 a :Class3 .
:vav3 a :TargetClass ;
    :prop :s3 .
"""


def test_087():
    res1 = validate(mixed_file_text, data_graph_format='turtle', shacl_graph_format='turtle', debug=True)
    conforms, graph, string = res1
    assert conforms



# -*- coding: utf-8 -*-
"""
https://www.w3.org/TR/shacl/#sparql-constraints
"""

from typing import Dict, List

import rdflib
from rdflib import URIRef

from pyshacl.constraints.constraint_component import ConstraintComponent
from pyshacl.consts import SH, SH_deactivated, SH_message, SH_select
from pyshacl.errors import ConstraintLoadError, ValidationFailure
from pyshacl.helper import get_query_helper_cls
from pyshacl.pytypes import GraphLike, SHACLExecutor
from pyshacl.shape import Shape

SH_sparql = SH.sparql
SH_SPARQLConstraintComponent = SH.SPARQLConstraintComponent


class SPARQLBasedConstraint(ConstraintComponent):
    """
    SHACL-SPARQL supports a constraint component that can be used to express restrictions based on a SPARQL SELECT query.
    Link:
    https://www.w3.org/TR/shacl/#sparql-constraints
    """

    shacl_constraint_component = SH_SPARQLConstraintComponent

    def __init__(self, shape: Shape) -> None:
        super(SPARQLBasedConstraint, self).__init__(shape)
        sg = self.shape.sg.graph
        sparql_node_list = set(self.shape.objects(SH_sparql))
        if len(sparql_node_list) < 1:
            raise ConstraintLoadError(
                "SPARQLConstraintComponent must have at least one sh:sparql predicate.",
                "https://www.w3.org/TR/shacl/#SPARQLConstraintComponent",
            )
        sparql_constraints = set()
        for s in iter(sparql_node_list):
            select_node_list = set(sg.objects(s, SH_select))
            if len(select_node_list) < 1:
                raise ConstraintLoadError(
                    "SPARQLConstraintComponent value for sh:select must have at least one sh:select predicate.",
                    "https://www.w3.org/TR/shacl/#SPARQLConstraintComponent",
                )
            elif len(select_node_list) > 1:
                raise ConstraintLoadError(
                    "SPARQLConstraintComponent value for sh:select must have at most one sh:select predicate.",
                    "https://www.w3.org/TR/shacl/#SPARQLConstraintComponent",
                )
            select_node = next(iter(select_node_list))
            if not (isinstance(select_node, rdflib.Literal) and isinstance(select_node.value, str)):
                raise ConstraintLoadError(
                    "SPARQLConstraintComponent value for sh:select must be a Literal with type xsd:string.",
                    "https://www.w3.org/TR/shacl/#SPARQLConstraintComponent",
                )
            message_node_list = set(sg.objects(s, SH_message))
            msgs = None
            if len(message_node_list) > 0:
                message = next(iter(message_node_list))
                if not (isinstance(message, rdflib.Literal) and isinstance(message.value, str)):
                    raise ConstraintLoadError(
                        "SPARQLConstraintComponent value for sh:message must be a Literal with type xsd:string.",
                        "https://www.w3.org/TR/shacl/#SPARQLConstraintComponent",
                    )
                msgs = message_node_list
            deactivated_node_list = set(sg.objects(s, SH_deactivated))
            deact = False
            if len(deactivated_node_list) > 0:
                deactivated = next(iter(deactivated_node_list))
                if not (isinstance(deactivated, rdflib.Literal) and isinstance(deactivated.value, bool)):
                    raise ConstraintLoadError(
                        "SPARQLConstraintComponent value for sh:deactivated must be a Literal with type xsd:boolean.",
                        "https://www.w3.org/TR/shacl/#SPARQLConstraintComponent",
                    )
                deact = bool(deactivated.value)
            SPARQLQueryHelper = get_query_helper_cls()
            query_helper = SPARQLQueryHelper(self.shape, s, select_node.value, messages=msgs, deactivated=deact)
            query_helper.collect_prefixes()
            sparql_constraints.add(query_helper)
        self.sparql_constraints = sparql_constraints

    @classmethod
    def constraint_parameters(cls) -> List[URIRef]:
        return [SH_sparql]

    @classmethod
    def constraint_name(cls) -> str:
        return "SPARQLConstraintComponent"

    def evaluate(
        self, executor: SHACLExecutor, target_graph: GraphLike, focus_value_nodes: Dict, _evaluation_path: List
    ):
        """
        :type executor: SHACLExecutor
        :type target_graph: rdflib.Graph
        :type focus_value_nodes: dict
        :type _evaluation_path: list
        """
        reports = []
        non_conformant = False

        for query_helper in self.sparql_constraints:
            if query_helper.deactivated:
                continue
            _nc, _r = self._evaluate_sparql_constraint(query_helper, target_graph, focus_value_nodes)
            non_conformant = non_conformant or _nc
            reports.extend(_r)
        return (not non_conformant), reports

    def _evaluate_sparql_constraint(self, sparql_constraint, target_graph, f_v_dict):
        reports = []
        non_conformant = False
        extra_messages = sparql_constraint.messages or None
        rept_kwargs = {'source_constraint': sparql_constraint.node, 'extra_messages': extra_messages}
        for f, value_nodes in f_v_dict.items():
            # we don't use value_nodes in the sparql constraint
            # All queries are done on the corresponding focus node.
            init_binds, sparql_text = sparql_constraint.pre_bind_variables(f)
            sparql_text = sparql_constraint.apply_prefixes(sparql_text)

            try:
                violating_vals = self._validate_sparql_query(sparql_text, init_binds, target_graph)

            except ValidationFailure as e:
                raise e
            if not self.shape.is_property_shape:
                result_val = f
            else:
                result_val = None
            for v in violating_vals:
                non_conformant = True
                if isinstance(v, bool) and v is True:
                    rept = self.make_v_result(target_graph, f, value_node=result_val, **rept_kwargs)
                elif isinstance(v, tuple):
                    if len(v) == 2:
                        (_, vars_dict) = v
                        rept = self.make_v_result(
                            target_graph, f, value_node=result_val, bound_vars=vars_dict, **rept_kwargs
                        )
                    elif len(v) == 4:
                        (t, p, v, vars_dict) = v
                        if v is None:
                            v = result_val
                        rept = self.make_v_result(
                            target_graph,
                            t or f,
                            value_node=v,
                            result_path=p,
                            bound_vars=(t, p, v, vars_dict),
                            **rept_kwargs,
                        )
                    else:
                        # len must equal 3
                        t, p, v = v
                        if v is None:
                            v = result_val
                        rept = self.make_v_result(
                            target_graph, t or f, value_node=v, result_path=p, bound_vars=(t, p, v), **rept_kwargs
                        )

                else:
                    rept = self.make_v_result(target_graph, f, value_node=v, **rept_kwargs)
                reports.append(rept)
        return non_conformant, reports

    def _validate_sparql_query(self, query, init_binds, target_graph):
        results = target_graph.query(query, initBindings=init_binds)
        if not results or len(results.bindings) < 1:
            return []
        violations = []
        dedup_set = set()
        for r in results:
            var_dict: Dict = r.asdict()
            f = var_dict.pop('failure', None)
            if f is not None:
                if True in dedup_set:
                    continue
                violations.append((True, var_dict))
                dedup_set.add(True)
            else:
                p = var_dict.pop('path', None)
                v = var_dict.pop('value', None)
                t = var_dict.pop('this', None)
                if (p is not None) or (v is not None) or (t is not None):
                    # Guard against duplicate results
                    if (t, p, v, var_dict) in violations:
                        continue
                    violations.append((t, p, v, var_dict))

        return violations



#
#
import typing
from typing import Dict

from rdflib import Literal

from pyshacl.consts import SH, SH_jsFunctionName, SH_jsLibrary
from pyshacl.errors import ConstraintLoadError

from .context import SHACLJSContext

if typing.TYPE_CHECKING:
    from pyshacl.shapes_graph import ShapesGraph

SH_jsLibraryURL = SH.jsLibraryURL


class JSExecutable(object):
    __slots__ = ("sg", "node", "fn_name", "libraries")

    def __new__(cls, shapes_graph: 'ShapesGraph', node):
        return super(JSExecutable, cls).__new__(cls)

    def __init__(self, shapes_graph: 'ShapesGraph', node):
        self.node = node
        self.sg = shapes_graph
        fn_names = set(shapes_graph.objects(node, SH_jsFunctionName))
        if len(fn_names) < 1:
            raise ConstraintLoadError(
                "At least one sh:jsFunctionName must be present on a JS Executable.",
                "https://www.w3.org/TR/shacl-js/#dfn-javascript-executables",
            )
        elif len(fn_names) > 1:
            raise ConstraintLoadError(
                "At most one sh:jsFunctionName can be present on a JS Executable.",
                "https://www.w3.org/TR/shacl-js/#dfn-javascript-executables",
            )
        fn_name = next(iter(fn_names))
        if not isinstance(fn_name, Literal):
            raise ConstraintLoadError(
                "sh:jsFunctionName must be an RDF Literal with type xsd:string.",
                "https://www.w3.org/TR/shacl-js/#dfn-javascript-executables",
            )
        else:
            fn_name = str(fn_name)
        self.fn_name = fn_name
        library_defs = shapes_graph.objects(node, SH_jsLibrary)
        seen_library_defs = []
        libraries: Dict = {}
        for libn in library_defs:
            # Library defs can only do two levels deep for now.
            # TODO: Make this recursive somehow to some further depth
            if libn in seen_library_defs:
                continue
            if isinstance(libn, Literal):
                raise ConstraintLoadError(
                    "sh:jsLibrary must not have a value that is a Literal.",
                    "https://www.w3.org/TR/shacl-js/#dfn-javascript-executables",
                )
            seen_library_defs.append(libn)
            jsLibraryURLs = list(shapes_graph.objects(libn, SH_jsLibraryURL))
            if len(jsLibraryURLs) > 0:
                libraries[libn] = libraries.get(libn, [])
            for u in jsLibraryURLs:
                if not isinstance(u, Literal):
                    raise ConstraintLoadError(
                        "sh:jsLibraryURL must have a value that is a Literal.",
                        "https://www.w3.org/TR/shacl-js/#dfn-javascript-executables",
                    )
                libraries[libn].append(str(u))
            library_defs2 = shapes_graph.objects(libn, SH_jsLibrary)
            for libn2 in library_defs2:
                if libn2 in seen_library_defs:
                    continue
                if isinstance(libn2, Literal):
                    raise ConstraintLoadError(
                        "sh:jsLibrary must not have a value that is a Literal.",
                        "https://www.w3.org/TR/shacl-js/#dfn-javascript-executables",
                    )
                seen_library_defs.append(libn2)
                jsLibraryURLs2 = list(shapes_graph.objects(libn2, SH_jsLibraryURL))
                if len(jsLibraryURLs2) > 0:
                    libraries[libn2] = libraries.get(libn2, [])
                for u2 in jsLibraryURLs2:
                    if not isinstance(u2, Literal):
                        raise ConstraintLoadError(
                            "sh:jsLibraryURL must have a value that is a Literal.",
                            "https://www.w3.org/TR/shacl-js/#dfn-javascript-executables",
                        )
                    libraries[libn2].append(str(u2))
        self.libraries = libraries

    def execute(self, data_graph, args_map, *args, mode=None, return_type=None, **kwargs):
        """
        :param data_graph:
        :param args_map:
        :param args:
        :param mode:
        :param return_type:
        :param kwargs:
        :return:
        :rtype: dict
        """
        if mode == "function":
            ctx = SHACLJSContext(data_graph, shapes_graph=None, **kwargs)
        else:
            ctx = SHACLJSContext(data_graph, shapes_graph=self.sg, **kwargs)

        for lib_node, lib_urls in self.libraries.items():
            for lib_url in lib_urls:
                ctx.load_js_library(lib_url)
        fn_args = ctx.get_fn_args(self.fn_name, args_map)
        rvals = ctx.run_js_function(self.fn_name, fn_args)
        res = rvals['_result']
        if mode == "function":
            rvals['_result'] = ctx.build_results_as_shacl_function(res, return_type)
        elif mode == "construct":
            rvals['_result'] = ctx.build_results_as_construct(res)
        elif mode == 'target':
            rvals['_result'] = ctx.build_results_as_target(res)
        else:
            rvals['_result'] = ctx.build_results_as_constraint(res)
        return rvals



# -*- coding: utf-8 -*-
#
"""
https://github.com/RDFLib/pySHACL/issues/286
"""
import sys

import rdflib

import pyshacl

shapes_data = '''\
@prefix sh: <http://www.w3.org/ns/shacl#>.
@prefix gx: <https://registry.lab.gaia-x.eu/development/api/trusted-shape-registry/v1/shapes/jsonld/trustframework#>.
@prefix xsd: <http://www.w3.org/2001/XMLSchema#>.

gx:LicenseShape a sh:NodeShape ;
    sh:property [
        sh:path gx:license ;
        sh:datatype xsd:string ;
        sh:in ("0BSD" "AAL" "EPL-2.0") ;
    ] ;
    sh:targetClass gx:License .
'''

data_shapes_in_xsd_string = '''\
@prefix sh: <http://www.w3.org/ns/shacl#>.
@prefix gx: <https://registry.lab.gaia-x.eu/development/api/trusted-shape-registry/v1/shapes/jsonld/trustframework#>.
@prefix xsd: <http://www.w3.org/2001/XMLSchema#>.

gx:LicenseShape a sh:NodeShape ;
    sh:property [
        sh:path gx:license ;
        sh:datatype xsd:string ;
        sh:in ("0BSD"^^xsd:string
                  "AAL"^^xsd:string
                  "EPL-2.0"^^xsd:string) ;
    ] ;
    sh:targetClass gx:License .
'''

data_g_text = '''\
{
    "@context": {
      "gx": "https://registry.lab.gaia-x.eu/development/api/trusted-shape-registry/v1/shapes/jsonld/trustframework#",
      "xsd": "http://www.w3.org/2001/XMLSchema#"
    },
    "@type": "gx:License",
    "gx:license": {
        "@value": "EPL-2.0",
        "@type": "xsd:string"
    }
}
'''


def test_286_a():
    shape_g = rdflib.Graph().parse(data=shapes_data, format='turtle')
    data_g = rdflib.Graph().parse(data=data_g_text, format="json-ld")
    conforms, results_graph, results_text = pyshacl.validate(
        data_g,
        shacl_graph=shape_g,
        debug=True,
        inference='none',
        advanced=False,
        meta_shacl=False,
    )
    assert not conforms
    assert "Results (1)" in results_text

def test_286_b():
    shape_g = rdflib.Graph().parse(data=data_shapes_in_xsd_string, format='turtle')
    data_g = rdflib.Graph().parse(data=data_g_text, format="json-ld")
    conforms, results_graph, results_text = pyshacl.validate(
        data_g,
        shacl_graph=shape_g,
        debug=True,
        inference='none',
        advanced=False,
        meta_shacl=False,
    )
    assert conforms

if __name__ == "__main__":
    test_286_b()
    sys.exit(test_286_b())



# -*- coding: utf-8 -*-
import itertools
from typing import TYPE_CHECKING, List, Optional, Sequence, Tuple, Union, cast

import rdflib

from pyshacl.consts import SH_object, SH_predicate, SH_subject
from pyshacl.errors import ReportableRuntimeError
from pyshacl.helper.expression_helper import nodes_from_node_expression
from pyshacl.rules.shacl_rule import SHACLRule

if TYPE_CHECKING:
    from rdflib.term import URIRef

    from pyshacl.pytypes import GraphLike, RDFNode, SHACLExecutor
    from pyshacl.shape import Shape

TRIPLE_RULE_ITERATE_LIMIT = 100


class TripleRule(SHACLRule):
    __slots__ = ("s", "p", "o")

    def __init__(self, executor: 'SHACLExecutor', shape: 'Shape', rule_node: 'rdflib.term.Identifier', **kwargs):
        """
        :param executor:
        :type executor: SHACLExecutor
        :param shape:
        :type shape: Shape
        :param rule_node:
        :type rule_node: rdflib.term.Identifier
        """
        super(TripleRule, self).__init__(executor, shape, rule_node, **kwargs)
        my_subject_nodes = set(self.shape.sg.objects(self.node, SH_subject))
        if len(my_subject_nodes) < 1:
            raise RuntimeError("No sh:subject")
        elif len(my_subject_nodes) > 1:
            raise RuntimeError("Too many sh:subject")
        self.s = next(iter(my_subject_nodes))

        my_predicate_nodes = set(self.shape.sg.objects(self.node, SH_predicate))
        if len(my_predicate_nodes) < 1:
            raise RuntimeError("No sh:predicate")
        elif len(my_predicate_nodes) > 1:
            raise RuntimeError("Too many sh:predicate")
        self.p = next(iter(my_predicate_nodes))

        my_object_nodes = set(self.shape.sg.objects(self.node, SH_object))
        if len(my_object_nodes) < 1:
            raise RuntimeError("No sh:object")
        elif len(my_object_nodes) > 1:
            raise RuntimeError("Too many sh:object")
        self.o = next(iter(my_object_nodes))

    def apply(
        self,
        data_graph: 'GraphLike',
        focus_nodes: Optional[Sequence['RDFNode']] = None,
        target_graph_identifier: Optional['URIRef'] = None,
    ) -> int:
        focus_list: Sequence['RDFNode']
        if focus_nodes is not None:
            focus_list = list(focus_nodes)
        else:
            focus_list = list(self.shape.focus_nodes(data_graph))
        if self.executor.focus_nodes is not None and len(self.executor.focus_nodes) > 0:
            filtered_focus_nodes: List[Union[rdflib.URIRef]] = []
            for _fo in focus_list:  # type: RDFNode
                if isinstance(_fo, rdflib.URIRef) and _fo in self.executor.focus_nodes:
                    filtered_focus_nodes.append(_fo)
            len_filtered_focus = len(filtered_focus_nodes)
            if len_filtered_focus < 1:
                return 0
            focus_list = filtered_focus_nodes
        # uses target nodes to find focus nodes
        applicable_nodes = self.filter_conditions(focus_list, data_graph)
        all_added = 0
        iterate_limit = int(TRIPLE_RULE_ITERATE_LIMIT)
        while True:
            if iterate_limit < 1:
                raise ReportableRuntimeError(
                    f"sh:rule iteration exceeded iteration limit of {TRIPLE_RULE_ITERATE_LIMIT}."
                )
            iterate_limit -= 1
            added = 0
            to_add = []
            for a in applicable_nodes:
                s_set = nodes_from_node_expression(self.s, a, data_graph, self.shape.sg)
                p_set = nodes_from_node_expression(self.p, a, data_graph, self.shape.sg)
                o_set = nodes_from_node_expression(self.o, a, data_graph, self.shape.sg)
                new_triples = itertools.product(s_set, p_set, o_set)
                this_added = False
                for i in iter(new_triples):
                    if not this_added and i not in data_graph:
                        this_added = True
                    to_add.append(i)
                if this_added:
                    added += 1
            if added > 0:
                if isinstance(data_graph, (rdflib.Dataset, rdflib.ConjunctiveGraph)):
                    if target_graph_identifier is not None:
                        target_graph = data_graph.get_context(target_graph_identifier)
                    else:
                        target_graph = data_graph.default_context
                else:
                    target_graph = data_graph
                for i in to_add:
                    target_graph.add(cast(Tuple['RDFNode', 'RDFNode', 'RDFNode'], i))
                all_added += added
                if self.iterate:
                    continue  # Jump up to iterate
                else:
                    break  # Don't iterate
            break
        return all_added



from pyshacl import validate
from rdflib import Dataset, Graph
from rdflib.plugins.stores.sparqlstore import SPARQLStore

# Remote sparql endpoint:
my_sparql_endpoint = "https://my.service.com/repo/data/sparql"
store = SPARQLStore(my_sparql_endpoint, auth=('username', 'password'))
data_dataset = Dataset(store, default_union=True)

# Shapes graph:
shapes_graph = Graph().parse("my_shapes.ttl", format="turtle")

# Validate:
result_tuple = validate(data_dataset, shacl_graph=shapes_graph, sparql_mode=True)
conforms, results_graph, results_text = result_tuple



from rdflib import Graph
from pyshacl import validate

shapes_graph = '''\
@prefix rdf: <http://www.w3.org/1999/02/22-rdf-syntax-ns#> .
@prefix rdfs: <http://www.w3.org/2000/01/rdf-schema#> .
@prefix sh: <http://www.w3.org/ns/shacl#> .
@prefix xsd: <http://www.w3.org/2001/XMLSchema#> .
@prefix ex: <http://example.com/ex#> .

ex:Rectangle
  rdf:type rdfs:Class ;
  rdf:type sh:NodeShape ;
  rdfs:label "Rectangle" ;
  rdfs:subClassOf rdfs:Resource ;
  sh:property [
      sh:path ex:height ;
      sh:datatype xsd:integer ;
      sh:maxCount 1 ;
      sh:minCount 1 ;
      sh:name "height" ;
    ] ;
  sh:property [
      sh:path ex:width ;
      sh:datatype xsd:integer ;
      sh:maxCount 1 ;
      sh:minCount 1 ;
      sh:name "width" ;
    ] ;
  sh:property [
      sh:path ex:area ;
      sh:datatype xsd:integer ;
      sh:maxCount 1 ;
      sh:minCount 1 ;
      sh:name "area" ;
    ] ;
.

ex:CheckArea
  rdf:type sh:PropertyShape ;
  sh:path ex:area ;
  sh:sparql ex:CheckArea-sparql ;
  sh:targetClass ex:Rectangle ;
.
ex:CheckArea-sparql
  rdf:type sh:SPARQLConstraintObject ;
  sh:message "Height * Width = Area." ;
  sh:prefixes <http://example.com/ex> ;
  sh:select """
		SELECT $this ?value
		WHERE {
		    $this ex:width ?width .
		    $this ex:height ?height .
			$this $PATH ?value .
			FILTER (ex:multiply(?width, ?height) != ?value)
		}
		""" ;
.
ex:multiply
  a sh:JSFunction ;
  rdfs:comment "Multiplies its two arguments $op1 and $op2." ;
  sh:parameter [
    sh:path ex:op1 ;
    sh:datatype xsd:integer ;
    sh:description "The first operand" ;
  ] ;
  sh:parameter [
    sh:path ex:op2 ;
    sh:datatype xsd:integer ;
    sh:description "The second operand" ;
  ] ;
  sh:returnType xsd:integer ;
  sh:jsLibrary [ sh:jsLibraryURL "file:./test/resources/js/multiply.js" ] ;
  sh:jsFunctionName "multiply" ;
.
'''

data_graph = '''\
@prefix rdf: <http://www.w3.org/1999/02/22-rdf-syntax-ns#> .
@prefix rdfs: <http://www.w3.org/2000/01/rdf-schema#> .
@prefix xsd: <http://www.w3.org/2001/XMLSchema#> .
@prefix ex: <http://example.com/ex#> .
@prefix exdata: <http://example.com/exdata#> .

exdata:NonSquareRectangle
  rdf:type ex:Rectangle ;
  ex:height 3 ;
  ex:width 4 ;
  ex:area 12 ;
.
exdata:SquareRectangle
  rdf:type ex:Rectangle ;
  ex:height 6 ;
  ex:width 6 ;
  ex:area 12 ;
.

'''


def test_js_function():
    s1 = Graph().parse(data=shapes_graph, format="turtle")
    g1 = Graph().parse(data=data_graph, format="turtle")
    conforms, result_graph, result_text = validate(g1, shacl_graph=s1, advanced=True, debug=True, js=True)
    assert not conforms


if __name__ == "__main__":
    test_js_function()



# HTTP Server for PySHACL
import os
import sys
from typing import Union

try:
    import sanic.application.logo
    from sanic import Request, Sanic
    from sanic.exceptions import InvalidUsage
    from sanic.response import HTTPResponse, JSONResponse, text
    from sanic_cors.extension import CORS
    from sanic_ext import Extend, openapi
except ImportError:
    raise RuntimeError("The optional PySHACL HTTP server components are not installed. See README.md for details.")

from dataclasses import dataclass
from enum import Enum
from textwrap import dedent

from sanic_ext.extensions.openapi import types as openapi_types
from sanic_ext.extensions.openapi.definitions import RequestBody, Response

from . import __version__ as pyshacl_version
from . import validate
from .errors import ConstraintLoadError, ReportableRuntimeError, RuleLoadError, ShapeLoadError, ValidationFailure

API_VERSION = "v1"


class InferenceKind(Enum):
    NONE = "none"
    RDFS = "rdfs"
    OWLRL = "owlrl"
    BOTH = "both"


class InputRDFFormat(Enum):
    AUTO = "auto"
    TURTLE = "turtle"
    XML = "xml"
    JSONLD = "json-ld"
    NT = "nt"
    N3 = "n3"


@dataclass
class ValidationRequest:
    data_graph: openapi_types.String(  # type: ignore[valid-type]
        required=True, title="DataGraph", description="The target DataGraph to validate (serialized in an RDF string)"
    )
    shapes_graph: openapi_types.String(  # type: ignore[valid-type]
        required=False,
        nullable=True,
        title="DataGraph",
        description="Your SHACL ShapesGraph (serialized in an RDF string)",
    )
    ontology_graph: openapi_types.String(  # type: ignore[valid-type]
        required=False, nullable=True, title="OntologyGraph", description="Optional ontological definitions graph"
    )
    data_graph_format: openapi_types.String(  # type: ignore[valid-type]
        required=False,
        nullable=False,
        default="auto",
        title="DataGraph RDF format",
        description="Optionally specify the RDF format for your DataGraph",
        enum=InputRDFFormat,
    )
    shapes_graph_format: openapi_types.String(  # type: ignore[valid-type]
        required=False,
        nullable=False,
        default="auto",
        title="ShapesGraph RDF format",
        description="Optionally specify the RDF format for your ShapesGraph",
        enum=InputRDFFormat,
    )
    ontology_graph_format: openapi_types.String(  # type: ignore[valid-type]
        required=False,
        nullable=False,
        default="auto",
        title="OntologyGraph RDF format",
        description="Optionally specify the RDF format for your OntologyGraph",
        enum=InputRDFFormat,
    )
    advanced: openapi_types.Boolean(  # type: ignore[valid-type]
        required=False,
        nullable=False,
        default=False,
        title="Advanced",
        description="Enable features from the SHACL Advanced Features spec.",
    )
    inference: openapi_types.String(  # type: ignore[valid-type]
        required=False,
        nullable=False,
        default="none",
        title="Pre-Inference",
        description="Set a pre-inference option",
        enum=InferenceKind,
    )
    do_owl_imports: openapi_types.Boolean(  # type: ignore[valid-type]
        required=False,
        nullable=False,
        default=False,
        title="Do OWL Imports",
        description="Enable the feature to follow links to import OWL ontologies in Shapes graph and Ontology Graph.",
    )
    allow_infos: openapi_types.Boolean(  # type: ignore[valid-type]
        required=False,
        nullable=False,
        default=False,
        title="Allow Infos",
        description="The datagraph will still be considered conformant when encountering constraint failures with "
        "sh:Info level severity.",
    )
    allow_warnings: openapi_types.Boolean(  # type: ignore[valid-type]
        required=False,
        nullable=False,
        default=False,
        title="Allow Warnings",
        description="The datagraph will still be considered conformant when encountering constraint failures with "
        "sh:Warning or sh:Info level severity.",
    )
    iterate_rules: openapi_types.Boolean(  # type: ignore[valid-type]
        required=False,
        nullable=False,
        default=False,
        title="Iterate Rules",
        description="Continue to execute SHACL Rules until the resulting output graph reaches steady state. "
        "This only works when advanced mode is enabled, and is usually not required.",
    )
    js: openapi_types.String(  # type: ignore[valid-type]
        required=False, nullable=False, default=False, title="JS", description="Enable SHACL-JS validator extension"
    )
    metashacl: openapi_types.Boolean(  # type: ignore[valid-type]
        required=False,
        nullable=False,
        default=False,
        title="MetaSHACL",
        description="Validate your SHACL Shapesfile against the shacl-shacl shapes before validating the datagraph.",
    )


failures_array_ref = openapi.Component(
    openapi_types.Array(
        items=openapi.String(title="ValidationFailure", required=True, nullable=False, format="text/plain"),
        required=False,
        nullable=True,
        description="Optional array of validation errors in text format.",
        format="application/json",
    ),
    name="ValidationFailureArray",
)


@dataclass
class ValidationResponseSimple:
    conforms: openapi_types.Boolean(  # type: ignore[valid-type]
        required=True, title="Conforms", description="The datagraph conforms to the SHACL Shapes"
    )
    validation_report: openapi_types.String(  # type: ignore[valid-type]
        required=False,
        nullable=True,
        title="ValidationReport",
        description="The text representation of the validation report.",
    )
    validation_failures: failures_array_ref  # type: ignore[valid-type]


def make_validation_response_RDF(format, mimetype):
    this_type_ref = openapi.Component(
        openapi_types.String(
            required=True,
            title=f"ValidationReport{format}",
            description=f"The {format} representation of the validation report.",
            format=mimetype,
        ),
        name=f"ValidationReport{format}",
    )
    return openapi.Component(
        openapi_types.Schema(oneOf=[this_type_ref, failures_array_ref], name=f"ValidationResponse{format}"),
        name=f"ValidationResponse{format}",
    )


validation_request_ref = openapi.Component(ValidationRequest)
validation_response_simple_ref = openapi.Component(ValidationResponseSimple)
validation_response_ntriples_ref = make_validation_response_RDF("NTriples", "application/n-triples")
validation_response_json_ld_ref = make_validation_response_RDF("JSONLD", "application/ld+json")
validation_response_xml_ref = make_validation_response_RDF("ApplicationXML", "application/xml")
validation_response_rdf_ref = make_validation_response_RDF("RDFXML", "application/rdf+xml")
validation_response_ttl_ref = make_validation_response_RDF("Turtle", "text/turtle")

ALLOWED_RESPONSE_TYPES = {
    "text/plain": validation_response_simple_ref,
    "application/json": validation_response_simple_ref,
    "application/ld+json": validation_response_json_ld_ref,
    "application/xml": validation_response_xml_ref,
    "application/rdf+xml": validation_response_rdf_ref,
    "text/turtle": validation_response_ttl_ref,
    "application/n-triples": validation_response_ntriples_ref,
}


@openapi.definition(
    summary="Validate",
    description="Send a validation request, consisting of a DataGraph, SHACL shapes graph, and optional parameters.",
    body=RequestBody(
        {"application/json": validation_request_ref},
        required=True,
        description="ValidationRequest body",
    ),
    validate=False,
    response=Response(ALLOWED_RESPONSE_TYPES, status=200),
)
async def sh_validate(request: Request) -> HTTPResponse:
    content_type = "application/json"  # Default content type is the fallback
    content_types = (request.headers.getall("Content-Type"),)
    for c_t in content_types:
        if isinstance(c_t, (list, tuple, set)):
            content_types_array = c_t
        else:
            content_types_array = [c_t]
        for c_t2 in content_types_array:
            split_ct = [p.strip() for p in c_t2.split(",")]
            for c_t3 in split_ct:
                content_type = ([p.strip() for p in c_t3.split(";")][0]).lower()

    accept_type = "text/plain"  # Default return type is the fallback
    accept_types = (request.headers.getall("Accept"),)
    for a_t in accept_types:
        if isinstance(a_t, (list, tuple, set)):
            accept_types_array = a_t
        else:
            accept_types_array = [a_t]
        for a_t2 in accept_types_array:
            split_at = [p.strip() for p in a_t2.split(",")]
            for a_t3 in split_at:
                accept_type = ([p.strip() for p in a_t3.split(";")][0]).lower()

    if content_type != "application/json":
        raise InvalidUsage(
            "Request should be encoded in format application/json in accordance with the OpenAPI schema."
        )
    if accept_type not in ALLOWED_RESPONSE_TYPES.keys():
        raise InvalidUsage("Invalid response type requested.")

    try:
        body = request.json
    except ValueError:
        raise InvalidUsage("Invalid JSON payload.")

    data_graph = body.get("data_graph", None)
    if data_graph is None:
        raise InvalidUsage("DataGraph was not provided.")
    data_graph_format = body.get("data_graph_format", None)
    shapes_graph = body.get("shapes_graph", None)
    shapes_graph_format = body.get("shapes_graph_format", None)
    ontology_graph = body.get("ontology_graph", None)
    ontology_graph_format = body.get("ontology_graph_format", None)
    advanced = body.get("advanced", False)
    inference = body.get("inference", 'none')
    do_owl_imports = body.get("do_owl_imports", False)
    allow_infos = body.get("allow_infos", False)
    allow_warnings = body.get("allow_warnings", False)
    iterate_rules = body.get("iterate_rules", False)
    js = body.get("js", False)
    metashacl = body.get("metashacl", False)

    if str(data_graph_format).lower() == "auto":
        data_graph_format = None
    if str(ontology_graph_format).lower() == "auto":
        ontology_graph_format = None
    if str(shapes_graph_format).lower() == "auto":
        shapes_graph_format = None
    if not advanced:
        iterate_rules = False

    try:
        _conforms, _graph, _text = validate(
            data_graph,
            shacl_graph=shapes_graph,
            ontology_graph=ontology_graph,
            data_graph_format=data_graph_format,
            shacl_graph_format=shapes_graph_format,
            ontology_graph_format=ontology_graph_format,
            advanced=advanced,
            inference=inference,
            do_owl_imports=do_owl_imports,
            js=js,
            metashacl=metashacl,
            allow_infos=allow_infos,
            allow_warnings=allow_warnings,
            iterate_rules=iterate_rules,
            debug=False,
        )
    except ValidationFailure as f:
        err = "Validation Failure: " + str(f)
        simple_resp = ValidationResponseSimple(conforms=None, validation_report=None, validation_failures=[err])
    except RuleLoadError as r:
        err = "Rule Load Error: " + str(r)
        simple_resp = ValidationResponseSimple(conforms=None, validation_report=None, validation_failures=[err])
    except ShapeLoadError as s:
        err = "Shape Load Error: " + str(s)
        simple_resp = ValidationResponseSimple(conforms=None, validation_report=None, validation_failures=[err])
    except ConstraintLoadError as c:
        err = "Constraint Load Error: " + str(c)
        simple_resp = ValidationResponseSimple(conforms=None, validation_report=None, validation_failures=[err])
    except ReportableRuntimeError as s:
        err = "Runtime Error: " + str(s)
        simple_resp = ValidationResponseSimple(conforms=None, validation_report=None, validation_failures=[err])
    except RuntimeError as e:
        err = "Runtime Error: Internal Error"
        import traceback

        # Print this to stderr in the console, not
        sys.stderr.write(f"{repr(e)}\r\n")
        traceback.print_tb(e.__traceback__, file=sys.stderr)
        sys.stderr.flush()
        simple_resp = ValidationResponseSimple(conforms=None, validation_report=None, validation_failures=[err])
    else:
        simple_resp = ValidationResponseSimple(conforms=_conforms, validation_report=_text, validation_failures=[])
    if accept_type == "text/plain":
        if simple_resp.validation_failures:
            failure_texts = "\r\n".join(simple_resp.validation_failures)
            resp_txt = f"validation_failures:\r\n{failure_texts}\r\n"
        else:
            resp_txt = f"conforms: {simple_resp.conforms}\r\n{simple_resp.validation_report}\r\n"
        return text(resp_txt, content_type=accept_type)
    elif accept_type == "application/json":
        resp_dict = {
            "conforms": simple_resp.conforms or False,
            "validation_report": simple_resp.validation_report,
            "validation_failures": simple_resp.validation_failures,
        }
        # Note, this is regular REST JSON, not a real JSON-LD RDF result
        return JSONResponse(resp_dict, content_type=accept_type)
    if simple_resp.validation_failures:
        # Not sure if we can translate these failures into RDF. Probably not possible.
        # Pull a swiftie and return JSON instead.
        resp_dict = {
            "conforms": False,
            "validation_report": simple_resp.validation_report,
            "validation_failures": simple_resp.validation_failures,
        }
        return JSONResponse(resp_dict, content_type="application/json")
    if accept_type == "application/ld+json":
        graph_str = _graph.serialize(format="json-ld")
    elif accept_type == "text/turtle":
        graph_str = _graph.serialize(format="turtle")
    elif accept_type == "application/xml" or accept_type == "application/rdf+xml":
        graph_str = _graph.serialize(format="xml")
    else:
        graph_str = _graph.serialize(format="ntriples")
    return HTTPResponse(body=graph_str, content_type=accept_type)


BASE_LOGO = """
            PySHACL
    Running HTTP REST Service

"""

COLOR_LOGO = """\033[48;2;255;13;104m                     \033[0m
\033[38;2;255;255;255;48;2;255;13;104m   ███████╗██╗  ██╗  \033[0m
\033[38;2;255;255;255;48;2;255;13;104m   ██╔════╝██║  ██║  \033[0m
\033[38;2;255;255;255;48;2;255;13;104m   ███████╗███████║  \033[0m
\033[38;2;255;255;255;48;2;255;13;104m   ╚════██║██╔══██║  \033[0m
\033[38;2;255;255;255;48;2;255;13;104m   ███████║██║  ██║  \033[0m
\033[38;2;255;255;255;48;2;255;13;104m   ╚══════╝╚═╝  ╚═╝  \033[0m
\033[48;2;255;13;104m   Running PySHACL   \033[0m
   HTTP REST Service """

FULL_COLOR_LOGO = """\033[38;2;255;13;104m             \033[0m
\033[38;2;255;13;104m ██████╗ ██╗   ██╗\033[0m ███████╗██╗  ██╗ █████╗  ██████╗██╗
\033[38;2;255;13;104m ██╔══██╗╚██╗ ██╔╝\033[0m ██╔════╝██║  ██║██╔══██╗██╔════╝██║
\033[38;2;255;13;104m ██████╔╝ ╚████╔╝ \033[0m ███████╗███████║███████║██║     ██║
\033[38;2;255;13;104m ██╔═══╝   ╚██╔╝  \033[0m ╚════██║██╔══██║██╔══██║██║     ██║
\033[38;2;255;13;104m ██║        ██║   \033[0m ███████║██║  ██║██║  ██║╚██████╗███████╗
\033[38;2;255;13;104m ╚═╝        ╚═╝   \033[0m ╚══════╝╚═╝  ╚═╝╚═╝  ╚═╝ ╚═════╝╚══════╝
"""  # noqa


def app_factory() -> Sanic:
    sanic.application.logo.BASE_LOGO = BASE_LOGO
    sanic.application.logo.COLOR_LOGO = COLOR_LOGO
    sanic.application.logo.FULL_COLOR_LOGO = FULL_COLOR_LOGO
    app = Sanic("PySHACL")
    app.config['MOTD_DISPLAY']["Application"] = "PySHACL HTTP SERVICE"
    CORS_OPTIONS = {"resources": r'/*', "origins": "*", "methods": ["GET", "POST", "HEAD", "OPTIONS"]}
    app.config['CORS_AUTOMATIC_OPTIONS'] = True
    # Disable sanic-ext built-in CORS, and add the Sanic-CORS plugin
    Extend(app, extensions=[CORS], config={"CORS": False, "CORS_OPTIONS": CORS_OPTIONS})

    app.ext.openapi.describe(
        "PySHACL Simple Validation Service API",
        version=API_VERSION,
        description=dedent(
            f"""
            # Info
            PySHACL is hosting its own simple validation service API endpoint.

            PySHACL version: `{pyshacl_version}`

            PySHACL API Spec Version: `{API_VERSION}`

            _Previously a standalone wrapper service, this feature is now built into PySHACL._
            """
        ),
    )
    app.route("/validate", methods=('POST', 'OPTIONS'))(sh_validate)
    return app


def run_server():
    PYSHACL_SERVER_LISTEN = os.getenv("PYSHACL_SERVER_LISTEN", "127.0.0.1")
    PYSHACL_SERVER_PORT_str = os.getenv("PYSHACL_SERVER_PORT", "8099")
    PYSHACL_SERVER_HOSTNAME = os.getenv("PYSHACL_SERVER_HOSTNAME", "")
    try:
        PYSHACL_SERVER_PORT = int(PYSHACL_SERVER_PORT_str)
    except ValueError:
        raise RuntimeError("Invalid port given for PySHACL Server Port")
    app: Sanic = app_factory()
    if PYSHACL_SERVER_HOSTNAME:
        app.config.SERVER_NAME = PYSHACL_SERVER_HOSTNAME
    app.run(
        PYSHACL_SERVER_LISTEN,
        port=PYSHACL_SERVER_PORT,
        access_log=False,
        auto_reload=False,
        single_process=True,
        legacy=False,
        debug=False,
    )


def main(prog: Union[None, str] = None) -> None:
    """CLI entrypoint for the HTTP Service"""
    try:
        run_server()
    except RuntimeError as r:
        import traceback

        sys.stderr.write(f"{r}\r\n")
        traceback.print_tb(r.__traceback__, file=sys.stderr)
        sys.stderr.flush()
        sys.exit(1)
    sys.exit(0)



# -*- coding: utf-8 -*-
#
"""
https://github.com/RDFLib/pySHACL/issues/227
"""
import sys

import rdflib

import pyshacl

shapes_data = '''\
@prefix ex-sh-rl: <http://otl.example.eu/ex/def/shape-rule/> .
@prefix rdfs: <http://www.w3.org/2000/01/rdf-schema#> .
@prefix sh: <http://www.w3.org/ns/shacl#> .
@prefix xsd: <http://www.w3.org/2001/XMLSchema#> .


ex-sh-rl:bothOrsAtOnce a sh:NodeShape ;
	sh:targetClass rdfs:Class ;
	sh:message "Wrong because neither object of sh:class or subject of sh:property" ;
	sh:or
		(
			[

				sh:path ( [ sh:inversePath [ sh:zeroOrMorePath rdfs:subClassOf ] ] sh:property ) ;
				sh:minCount 1 ;
			]
			[
				sh:path ( [ sh:inversePath [ sh:zeroOrMorePath rdfs:subClassOf ] ] [sh:inversePath sh:class ] ) ;
				sh:minCount 1 ;
			]
		) .



ex-sh-rl:Or1 a sh:NodeShape ;
	sh:targetClass rdfs:Class ;
	sh:message "Wrong because not subject sh:property" ;
	sh:property [
		sh:path ( [ sh:inversePath [ sh:zeroOrMorePath rdfs:subClassOf ] ] sh:property ) ;
				sh:minCount 1 ;
	].

ex-sh-rl:Or2 a sh:NodeShape ;
	sh:targetClass rdfs:Class ;
	sh:message "Wrong because not object of sh:class" ;
	sh:property [
				sh:path ( [ sh:inversePath [ sh:zeroOrMorePath rdfs:subClassOf ] ] [sh:inversePath sh:class ] ) ;
				sh:minCount 1 ;
			]
		.
'''

data_g_text = '''\
@prefix ex-sh-rl: <http://otl.example.eu/ex/def/shape-rule/> .
@prefix rdfs: <http://www.w3.org/2000/01/rdf-schema#> .
@prefix sh: <http://www.w3.org/ns/shacl#> .
@prefix xsd: <http://www.w3.org/2001/XMLSchema#> .


ex-sh-rl:CorrectSuperClass a rdfs:Class.
ex-sh-rl:CorrectClass a rdfs:Class ;
	rdfs:subClassOf ex-sh-rl:CorrectSuperClass.

ex-sh-rl:something sh:class ex-sh-rl:CorrectClass.

ex-sh-rl:WrongClass a rdfs:Class .
'''


def test_227():
    shape_g = rdflib.Graph().parse(data=shapes_data, format='turtle')
    data_g = rdflib.Graph().parse(data=data_g_text, format="turtle")
    conforms, results_graph, results_text = pyshacl.validate(
        data_g,
        shacl_graph=shape_g,
        debug=True,
        inference='none',
        advanced=False,
        meta_shacl=False,
    )
    assert not conforms


if __name__ == "__main__":
    sys.exit(test_227())



# -*- coding: utf-8 -*-
"""
https://github.com/RDFLib/pySHACL/issues/146
"""

import warnings


def test_146() -> None:
    # Ensure that importing pyshacl triggers no warnings.
    with warnings.catch_warnings(record=True) as warning_context:
        # Cause all warnings to always be triggered.
        warnings.simplefilter("always")
        # Import pyshacl, which should not trigger any warnings
        import pyshacl

        # Verify some things
        assert len(warning_context) == 0


if __name__ == "__main__":
    test_146()



# -*- coding: utf-8 -*-
#
import http.client
import os
import pickle
import platform
import sys
from io import BufferedIOBase, BytesIO, TextIOBase, UnsupportedOperation
from logging import WARNING, Logger, getLogger
from pathlib import Path, PurePath
from typing import IO, List, Optional, Union, cast
from urllib import request
from urllib.error import HTTPError
from urllib.parse import unquote_to_bytes

import rdflib
from rdflib.namespace import SDO, NamespaceManager
from rdflib.term import URIRef

from .clone import clone_dataset, clone_graph

SCHEMA = SDO

ConjunctiveLike = Union[rdflib.ConjunctiveGraph, rdflib.Dataset]
GraphLike = Union[ConjunctiveLike, rdflib.Graph]

is_windows = platform.system() == "Windows"
MAX_OWL_IMPORT_DEPTH = 3
baked_in = {}


def path_from_uri(uri: Union[str, URIRef], relative_to: Union[PurePath, None] = None) -> Path:
    """Return a new path from the given 'file' URI."""
    if not uri.startswith('file:'):
        raise ValueError(f"File URI does not start with 'file:': {uri!r}")
    path: str = str(uri)[5:]
    if path[:3] == '///':
        # Remove empty authority
        path = path[2:]
    elif path[:12] == '//localhost/':
        # Remove 'localhost' authority
        path = path[11:]
    if path[:3] == '///' or (path[:1] == '/' and path[2:3] in ':|'):
        # Remove slash before DOS device/UNC path
        path = path[1:]
    if path[1:2] == '|':
        # Replace bar with colon in DOS drive
        path = path[:1] + ':' + path[2:]
    path_ret = Path(os.fsdecode(unquote_to_bytes(path)))
    if relative_to is None:
        return path_ret
    else:
        return Path(relative_to).joinpath(path_ret)


def add_baked_in(url, graph_path):
    baked_in[url] = graph_path


def get_rdf_from_web(url: Union[URIRef, str]):
    """

    :param url:
    :type url: rdflib.URIRef | str
    :return:
    """
    no_hash_url: str = str(url).rstrip("#")
    kind: Optional[str]
    if no_hash_url in baked_in:
        g = baked_in[no_hash_url]
        if isinstance(g, str):
            if g[-7:] == ".pickle":
                with open(g, 'rb') as g_pickle:
                    u = pickle.Unpickler(g_pickle, fix_imports=False)
                    g_store, identifier = u.load()
                graph = rdflib.Graph(store=g_store, identifier=identifier)
                kind = "graph"
            else:
                graph = rdflib.Graph()
                graph.parse(g)
                kind = None
        else:
            graph = g
            kind = None
        return graph, None, kind, False

    # Ask for everything we know about
    headers = {
        'Accept': 'text/turtle, application/rdf+xml, application/ld+json, application/n-triples, text/plain',
        'Accept-Encoding': 'identity',
    }
    known_format = None

    r: request.Request = request.Request(url, headers=headers)
    resp: http.client.HTTPResponse = request.urlopen(r)
    code: int = resp.getcode()
    if not (200 <= code <= 210):
        raise RuntimeError("Cannot pull RDF URL from the web: {}, code: {}".format(url, str(code)))

    filename = None
    content_dispositions: List[str] = resp.headers.get_all("Content-Disposition", [])
    for c_d in content_dispositions:
        cd_parts = [s.strip() for s in str(c_d).split(',')]
        for cd_part in cd_parts:
            if "filename=" in cd_part:
                filename = [f.strip() for f in str(cd_part).rsplit('filename=')][-1]
    if filename is None:
        try:
            filename = resp.geturl()
        except Exception:
            pass

    content_types: List[str] = resp.headers.get_all('Content-Type', [])
    for content_type in content_types:
        ct_parts = [s.strip() for s in str(content_type).split(',')]
        for ct_part in ct_parts:
            if ct_part.startswith("application/octet-stream"):
                known_format = 'auto'
            elif ct_part.startswith("text/turtle"):
                known_format = "turtle"
            elif ct_part.startswith("application/rdf+xml"):
                known_format = "xml"
            elif ct_part.startswith("application/xml"):
                known_format = "xml"
            elif ct_part.startswith("application/ld+json"):
                known_format = "json-ld"
            elif ct_part.startswith("application/n-triples"):
                known_format = "nt"
            else:
                continue
            break

    transfer_encodings: List[str] = resp.headers.get_all('Transfer-Encoding', [])
    for t_e in transfer_encodings:
        te_parts = [s.strip() for s in str(t_e).split(',')]
        for check in ('chunked', 'compress', 'deflate', 'gzip', 'x-gzip'):
            if check in te_parts:
                return resp, filename, known_format, False

    return resp, filename, known_format, True


# What's the difference between PublicID and BaseURI?
# The BaseURI is a part of Turtle and SPARQL spec, it is used to resolve relative URIs.
# The BaseURI usually ends with a filename (eg, https://example.com/validators/shapes)
# BaseURI can sometimes end with a / if URIs are relative to a directory.
# You will rarely see a BaseURI with a # on the end.
# The PublicID is a concept inherited from the XML specificaion
# RDFLib uses PublicID for the Identifier of a Graph. It is the canonical name of the graph,
# regardless of its hosted location. It is also used to refer to a Named Graph in a Dataset
# and this is the name referenced in the owl:imports [ schema:name <publicID> ] statement.
# PublicID is not found in the Turtle file, it is known outside the file only.
# PublicID can end with a / or a # if you want consistency with the graph's base prefix.
# Alternatively, PublicID may not have a symbol at the end.
# Note, PublicID is now called "Identifier" in the load_from_source function.


def load_from_source(
    source: Union[GraphLike, BufferedIOBase, TextIOBase, str, bytes],
    g: Optional[GraphLike] = None,
    rdf_format: Optional[str] = None,
    identifier: Optional[Union[URIRef, str]] = None,
    multigraph: bool = False,
    do_owl_imports: Union[bool, int] = False,
    import_chain: Optional[List[Union[URIRef, str]]] = None,
    logger: Optional[Logger] = None,
):
    """

    :param source:
    :param g: The Graph to load into, optional. If not given, a new Dataset or Graph will be created.
    :type g: rdflib.Graph | None
    :param rdf_format:
    :type rdf_format: str | None
    :param multigraph:
    :type multigraph: bool
    :param identifier: Identifier for the Named Graph being loaded. formerly "public_id"
    :type identifier: str | URIRef | None
    :param do_owl_imports:
    :type do_owl_imports: bool|int
    :param import_chain:
    :type import_chain: list | None
    :param logger:
    :type logger: Logger | None
    :return:
    """
    source_is_graph = False
    open_source: Optional[BufferedIOBase] = None
    source_was_open: bool = False
    source_as_file: Optional[BufferedIOBase] = None
    source_as_filename: Optional[str] = None
    source_as_bytes: Optional[bytes] = None
    filename = None
    # (Note, identifier is always first converted to a str, even if it is passes in as a URIRef)
    identifier = None if identifier is None else str(identifier)  # This is our passed-in id (formerly public_id)
    _maybe_id: Optional[str] = None  # For default-graph identifier
    base_uri: Optional[str] = None  # Base URI for relative URIs
    uri_prefix = None  # URI Prefix to bind to public ID
    if logger is None:
        logger = getLogger("rdfutil.load")
        logger.setLevel(WARNING)
    is_imported_graph = do_owl_imports and isinstance(do_owl_imports, int) and do_owl_imports > 1
    if isinstance(source, (rdflib.Graph, rdflib.ConjunctiveGraph, rdflib.Dataset)):
        source_is_graph = True
        if g is None:
            g = source
        else:
            raise RuntimeError(
                "Cannot pass in both source=rdflib.Graph/Dataset and g=graph. "
                "Source and dest cannot be the same graph."
            )
    elif isinstance(source, (BufferedIOBase, TextIOBase)):
        if hasattr(source, 'name'):
            filename = source.name  # type: ignore
            # Note, we _could_ preserve the relative file URI for the BaseURI
            # but the Turtle parser converts a relative BaseURI to absolute anyway
            # So just make it absolute here so we have a consistent reference to it.
            file_uri = Path(filename).absolute().as_uri()
            _maybe_id = file_uri
            base_uri = file_uri
        if isinstance(source, TextIOBase):
            buf = getattr(source, "buffer")  # type: BufferedIOBase
            source_as_file = source = buf
        else:
            source_as_file = source
        if hasattr(source, 'closed'):
            if not bool(source.closed):
                open_source = source
                source_was_open = True
        else:
            # Assume it is open now and it was open when we started.
            open_source = source
            source_was_open = True

    elif isinstance(source, str):
        if source == "stdin" or source == "-" or source == "/dev/stdin":
            _maybe_id = "/dev/stdin"
            # Don't set base_uri, it is not used for /dev/stdin
            filename = "/dev/stdin"
            source_as_filename = filename
        if source.startswith('file:'):
            # Make this abosolute, because the Turtle parser converts a relative BaseURI to absolute anyway
            # So just make it absolute here so we have a consistent reference to it.
            _path = path_from_uri(source, relative_to=None).absolute()
            filename = str(_path)
            _maybe_id = base_uri = _path.as_uri()
            source_as_filename = filename
        elif source.startswith('http:') or source.startswith('https:'):
            # It can be tricky to guess public_id from a web URL.
            # In this case we will always simply use the URL as the public_id as given.
            _maybe_id = source
            base_uri = source
            try:
                resp, resp_filename, web_format, raw_fp = get_rdf_from_web(source)
            except HTTPError:
                if is_imported_graph:
                    return g
                else:
                    raise
            if web_format == 'graph':
                source = resp
                source_is_graph = True
            elif web_format in ('auto', None):
                if resp_filename:
                    filename = resp_filename
                source_was_open = False
                source = open_source = resp
            else:
                rdf_format = web_format
                filename = resp_filename
                fp = resp.fp if raw_fp else resp
                source_was_open = False
                source = open_source = fp
        else:
            first_char = source[0]
            if is_windows and (first_char == '\\' or (len(source) > 3 and source[1:3] == ":\\")):
                filename = source
                source_as_filename = filename
            elif first_char == '/' or (len(source) > 2 and source[0:2] == "./"):
                filename = source
                source_as_filename = filename
            elif (
                first_char == '#'
                or first_char == '@'
                or first_char == '<'
                or first_char == '\n'
                or first_char == '{'
                or first_char == '['
            ):
                # Contains some JSON or XML or Turtle chars, it's not a path
                source_as_file = None
                source_as_filename = None
            elif len(source) >= 32 and '\n' in source[:32]:
                # Contains a new line near the start of the file, can't be a path
                source_as_file = None
                source_as_filename = None
            elif len(source) < 140:
                filename = source
                source_as_filename = filename
        if source_as_filename and filename:
            pid = os.getpid()
            fd0 = "/proc/{}/fd/0".format(str(pid))
            if filename == "/dev/stdin" or filename == fd0:
                source = source_as_file = open_source = cast(BufferedIOBase, sys.stdin.buffer)
                source_was_open = True
            else:
                try:
                    filename = os.readlink(filename)
                    if filename == fd0 or filename == "/dev/stdin":
                        source = source_as_file = open_source = cast(BufferedIOBase, sys.stdin.buffer)
                        source_was_open = True
                except OSError:
                    pass

        if not source_as_file and not source_as_filename and not open_source and isinstance(source, str):
            # source is raw RDF data.
            source_as_bytes = source = source.encode('utf-8')
    elif isinstance(source, bytes):
        if source.startswith(b'file:') or source.startswith(b'http:') or source.startswith(b'https:'):
            raise ValueError("file: and http: strings should be given as str, not bytes.")
        first_char_b: bytes = source[0:1]
        if (
            first_char_b == b'#'
            or first_char_b == b'@'
            or first_char_b == b'<'
            or first_char_b == b'\n'
            or first_char_b == b'{'
            or first_char_b == b'['
        ):
            # Contains some JSON or XML or Turtle stuff
            source_as_file = None
            source_as_filename = None
        elif len(source) < 140:
            filename = source.decode('utf-8')
            source_as_filename = filename
        if not source_as_file and not source_as_filename and not open_source:
            source_as_bytes = source
    else:
        raise ValueError("Cannot determine the format of the input graph")
    if g is None:
        if source_is_graph:
            target_g: Union[rdflib.Graph, rdflib.ConjunctiveGraph, rdflib.Dataset] = source  # type: ignore
        else:
            default_graph_base: Union[str, None] = base_uri if base_uri else (identifier if identifier else None)
            if multigraph:
                target_ds = rdflib.Dataset(default_graph_base=default_graph_base, default_union=True)
                target_ds.namespace_manager = NamespaceManager(target_ds, 'core')
                if identifier:  # if identifier is explicitly given, use that as a new named graph id
                    old_default_context = target_ds.default_context
                    if str(old_default_context.identifier) != identifier:
                        named_g = target_ds.graph(URIRef(identifier))
                        named_g.base = default_graph_base
                        target_ds.default_context = named_g
                        target_ds.remove_graph(old_default_context)
                else:
                    target_ds.default_context.namespace_manager = target_ds.namespace_manager
                    default_g = target_ds.default_context
                    target_ds.graph(default_g)
                target_g = target_ds
            else:
                target_g = rdflib.Graph(
                    bind_namespaces='core',
                    base=default_graph_base,
                    identifier=None if not identifier else URIRef(identifier),
                )

    else:
        if not isinstance(g, (rdflib.Graph, rdflib.Dataset, rdflib.ConjunctiveGraph)):
            raise RuntimeError("Passing in 'g' must be a rdflib Graph or Dataset.")
        target_g = g

    if filename and not rdf_format:
        if filename.endswith('.ttl'):
            rdf_format = rdf_format or 'turtle'
        elif filename.endswith('.nt'):
            rdf_format = rdf_format or 'nt'
        elif filename.endswith('.n3'):
            rdf_format = rdf_format or 'n3'
        elif filename.endswith('.json'):
            rdf_format = rdf_format or 'json-ld'
        elif filename.endswith('.nq') or filename.endswith('.nquads'):
            rdf_format = rdf_format or 'nquads'
        elif filename.endswith('.trig'):
            rdf_format = rdf_format or 'trig'
        elif filename.endswith('.xml') or filename.endswith('.rdf'):
            rdf_format = rdf_format or 'xml'
        elif filename.endswith('.hext'):
            rdf_format = rdf_format or 'hext'
    if source_as_filename and filename is not None and not open_source:
        filename = str(Path(filename).resolve())
        if not _maybe_id:
            _maybe_id = Path(filename).as_uri()
        source = open_source = cast(BufferedIOBase, open(filename, mode='rb'))
    if not open_source and source_as_bytes:
        source = open_source = BytesIO(source_as_bytes)  # type: ignore

    if open_source:
        _source = open_source
        # Check if we can seek
        try:
            _source.seek(0)  # type: ignore
        except (AttributeError, ValueError, UnsupportedOperation):
            # Read it all into memory
            new_bytes = BytesIO(_source.read())
            if not source_was_open:
                _source.close()
            source = _source = new_bytes
            source_was_open = False
        if rdf_format is None:
            line: Union[bytes, None] = _source.readline()
            line = None if line is None else line.lstrip()
            line_len: int = len(line) if line is not None else 0
            while line is not None and line_len == 0:
                line = _source.readline()
                line = None if line is None else line.lstrip()
                line_len = len(line) if line is not None else 0
            if line is not None:
                if line_len > 15:
                    line = line[:15]
                line = line.lower()
                if line.startswith(b"<!doctype html") or line.startswith(b"<html"):
                    raise RuntimeError("Attempted to load a HTML document as RDF.")
                if line.startswith(b"<?xml") or line.startswith(b"<xml") or line.startswith(b"<rdf:"):
                    rdf_format = "xml"
                if (
                    line.startswith(b"@prefix ")
                    or line.startswith(b"PREFIX ")
                    or line.startswith(b"@base ")
                    or line.startswith(b"# baseURI:")
                ):
                    rdf_format = "turtle"
            try:
                _source.seek(0)
            except (AttributeError, UnsupportedOperation):
                raise RuntimeError("Seek failed while identifying file type.")
            except ValueError:
                raise RuntimeError("File closed while identifying file type.")
        if rdf_format == 'turtle' or rdf_format == 'n3':
            # SHACL Shapes files and Data files can have extra RDF Metadata in the
            # Top header block, including #BaseURI and #Prefix.
            # The @base line is not read here, but it is parsed in the n3 parser
            while True:
                try:
                    line = _source.readline()
                    assert line is not None and len(line) > 0
                except AssertionError:
                    break
                # Strip line from start
                while len(line) > 0 and line[0:1] in b' \t\n\r\x0b\x0c\x85\xa0':
                    line = line[1:]
                # We reached the end of the line, check the next line
                if len(line) < 1:
                    continue
                # If this is not a comment, then this is the first non-comment line, we're done.
                if not line[0:1] == b'#':
                    break
                # Strip from start again, but now removing hashes too.
                while len(line) > 0 and line[0:1] in b'# \t\xa0':
                    line = line[1:]
                # Strip line from end
                while len(line) > 0 and line[-1:] in b' \t\n\r\x0b\x0c\x85\xa0':
                    line = line[:-1]
                spl = line.split(b':', 1)
                if len(spl) < 2:
                    continue
                keyword = spl[0].lower()
                # Strip keyword end
                while len(keyword) > 0 and keyword[-1:] in b' \t\n\r\x0b\x0c\x85\xa0':
                    keyword = keyword[:-1]
                if len(keyword) < 1:
                    continue
                wordval = spl[1]
                # Strip wordval start
                while len(wordval) > 0 and wordval[0:1] in b' \t\n\r\x0b\x0c\x85\xa0':
                    wordval = wordval[1:]
                if len(wordval) < 1:
                    continue
                wordval_str = wordval.decode('utf-8')
                if keyword == b"baseuri" or keyword == b"base":  # BASE is the SPARQL version of BaseURI
                    base_uri = wordval_str
                elif keyword == b"prefix":
                    uri_prefix = wordval_str
            try:
                _source.seek(0)
            except (AttributeError, UnsupportedOperation):
                raise RuntimeError("Seek failed while pre-parsing Turtle File.")
            except ValueError:
                raise RuntimeError("File closed while pre-parsing Turtle File.")

        # use base_uri if it is set, otherwise use identifier or _maybe_id
        parser_base_uri: Union[str, None] = base_uri if base_uri else (identifier if identifier else _maybe_id)
        if isinstance(target_g, (rdflib.Dataset, rdflib.ConjunctiveGraph)):
            if identifier:
                dest_g = target_g.get_context(URIRef(identifier))
                dest_g.base = parser_base_uri
            else:
                dest_g = target_g.default_context
                dest_g.base = parser_base_uri
            # parsing uses base_uri as the public_id, because it is used for relative URIs
            dest_g.parse(source=cast(IO[bytes], _source), format=rdf_format, publicID=parser_base_uri)
        else:
            target_g.parse(source=cast(IO[bytes], _source), format=rdf_format, publicID=parser_base_uri)
        # If the target was open to begin with, leave it open.
        if not source_was_open:
            _source.close()
        elif hasattr(_source, 'seek'):
            try:
                _source.seek(0)
            except (AttributeError, UnsupportedOperation):
                pass
            except ValueError:
                # The parser closed our file!
                pass
        source_is_graph = True
    elif source_is_graph and (target_g != source):
        # clone source into g
        if isinstance(target_g, (rdflib.Dataset, rdflib.ConjunctiveGraph)) and isinstance(
            source, (rdflib.Dataset, rdflib.ConjunctiveGraph)
        ):
            clone_dataset(source, target_g)
        elif isinstance(target_g, rdflib.Graph) and isinstance(source, (rdflib.Dataset, rdflib.ConjunctiveGraph)):
            raise RuntimeError("Cannot load a Dataset source into a Graph target.")
        elif isinstance(target_g, (rdflib.Dataset, rdflib.ConjunctiveGraph)) and isinstance(source, rdflib.Graph):
            _temp_target = rdflib.Graph(
                store=target_g.store,
                identifier=source.identifier if not identifier else URIRef(identifier),
                base=source.base if not base_uri else base_uri,
                namespace_manager=target_g.namespace_manager,
            )
            clone_graph(source, _temp_target)
        elif isinstance(target_g, rdflib.Graph) and isinstance(source, rdflib.Graph):
            clone_graph(source, target_g)
        else:
            raise RuntimeError("Cannot merge source graph into target graph.")

    if not source_is_graph:
        raise RuntimeError("Error opening graph from source.")

    if identifier:
        identifier_namespace = (
            identifier if (identifier.endswith('#') or identifier.endswith('/')) else identifier + "#"
        )
        if uri_prefix:
            if is_imported_graph and uri_prefix == '':
                # Don't reassign blank prefix, when importing subgraph
                pass
            else:
                has_named_prefix = target_g.store.namespace(uri_prefix)
                if not has_named_prefix:
                    target_g.namespace_manager.bind(uri_prefix, identifier_namespace)
        elif not is_imported_graph:
            existing_blank_prefix = target_g.store.namespace('')
            if not existing_blank_prefix:
                target_g.namespace_manager.bind('', identifier_namespace)
    if do_owl_imports:
        if isinstance(do_owl_imports, bool):
            do_owl_imports = 1 if do_owl_imports else 0
        elif isinstance(do_owl_imports, int):
            if do_owl_imports > MAX_OWL_IMPORT_DEPTH:
                return target_g
        else:
            do_owl_imports = 1

        if import_chain is None:
            import_chain = []
        if isinstance(target_g, (rdflib.Dataset, rdflib.ConjunctiveGraph)):
            if identifier:
                dest_g = target_g.get_context(URIRef(identifier))
            else:
                dest_g = target_g.default_context
        else:
            dest_g = target_g
        return chain_load_owl_imports(
            cast(Union[URIRef, rdflib.BNode], dest_g.identifier),
            dest_g.base,
            target_g,
            import_chain,
            do_owl_imports,
            multigraph,
        )
    return target_g


def chain_load_owl_imports(
    graph_id: Union[URIRef, rdflib.BNode, None],
    graph_base: Union[str, None],
    target_g: GraphLike,
    import_chain: List[Union[URIRef, str]],
    load_iter: int,
    multigraph: bool,
) -> GraphLike:
    if graph_base and (graph_base.endswith('#') or graph_base.endswith('/')):
        root_id: Union[URIRef, None] = URIRef(graph_base[:-1])
    else:
        root_id = URIRef(graph_base) if graph_base else None
    done_imports = 0

    def _load_from_imports_nodes(imports_nodes: List[Union[URIRef, rdflib.BNode]]) -> int:
        nonlocal target_g, multigraph, import_chain, load_iter
        _done_imports = 0
        for _i in imports_nodes:
            import_with_identifier: Union[str, None] = None
            if isinstance(_i, rdflib.BNode):
                urls = list(target_g.objects(_i, SCHEMA.url))
                prioritized_urls = []  # Tuples of (priority, url_str)
                # Value of type variable "SupportsRichComparisonT" of "sorted" cannot be "Node"
                # Maybe we need to add "SupportsRichComparisonT" to Node in RDFLib?
                for url_i in sorted(urls):  # type: ignore[type-var]
                    url_str = str(url_i)
                    if url_str.startswith("file:"):
                        prioritized_urls.append((1, url_str))
                    else:
                        prioritized_urls.append((9, url_str))
                _prio, imp_str = sorted(prioritized_urls)[0]  # this causes the first (alphabetically) URL to be used
                use_identifiers = list(target_g.objects(_i, SCHEMA.identifier))
                if len(use_identifiers) > 0:
                    import_with_identifier = str(next(iter(use_identifiers)))
            else:
                imp_str = str(_i)
            if imp_str in import_chain:
                continue
            if imp_str.startswith('file:'):
                imp_str = str(path_from_uri(imp_str, relative_to=None))
            load_from_source(
                imp_str,
                g=target_g,
                identifier=import_with_identifier,
                multigraph=multigraph,
                do_owl_imports=load_iter + 1,
                import_chain=import_chain,
            )
            _done_imports += 1
        return _done_imports

    if isinstance(target_g, (rdflib.ConjunctiveGraph, rdflib.Dataset)):
        # Don't care about named graphs, search across the whole
        # thing at once.
        target_g.default_union = True

    if root_id is not None:
        owl_imports = list(target_g.objects(root_id, rdflib.OWL.imports))
        if len(owl_imports) > 0:
            import_chain.append(str(root_id))
            _done_imports = _load_from_imports_nodes(owl_imports)  # type: ignore[arg-type]
            if _done_imports < 1:
                import_chain.pop()
            else:
                done_imports += _done_imports
    if done_imports < 1 and graph_base is not None and str(root_id) != graph_base:
        owl_imports = list(target_g.objects(URIRef(graph_base), rdflib.OWL.imports))
        if len(owl_imports) > 0:
            import_chain.append(graph_base)
            _done_imports = _load_from_imports_nodes(owl_imports)  # type: ignore[arg-type]
            if _done_imports < 1:
                import_chain.pop()
            else:
                done_imports += _done_imports
    if done_imports < 1 and graph_id is not None and root_id != graph_id:
        owl_imports = list(target_g.objects(graph_id, rdflib.OWL.imports))
        if len(owl_imports) > 0:
            import_chain.append(str(graph_id))
            _done_imports = _load_from_imports_nodes(owl_imports)  # type: ignore[arg-type]
            if _done_imports < 1:
                import_chain.pop()
            else:
                done_imports += _done_imports
    if done_imports < 1:
        ontologies = target_g.subjects(rdflib.RDF.type, rdflib.OWL.Ontology)
        for ont in ontologies:
            if ont == root_id or ont == graph_id:
                continue
            ont_str = str(ont)
            if ont_str == graph_base or ont_str in import_chain:
                continue
            owl_imports = list(target_g.objects(ont, rdflib.OWL.imports))
            if len(owl_imports) > 0:
                import_chain.append(ont_str)
                _done_imports = _load_from_imports_nodes(owl_imports)  # type: ignore[arg-type]
                if _done_imports < 1:
                    import_chain.pop()
                else:
                    done_imports += _done_imports
    return target_g



# -*- coding: utf-8 -*-
#

# This software was developed at the National Institute of Standards
# and Technology by employees of the Federal Government in the course
# of their official duties. Pursuant to title 17 Section 105 of the
# United States Code this software is not subject to copyright
# protection and is in the public domain. NIST assumes no
# responsibility whatsoever for its use by other parties, and makes
# no guarantees, expressed or implied, about its quality,
# reliability, or any other characteristic.
#
# We would appreciate acknowledgement if the software is used.

"""
https://github.com/RDFLib/pySHACL/issues/126
"""

from typing import Optional

from pyshacl import validate
from rdflib import Namespace, SH, URIRef

NS_EX = Namespace("http://example.org/ns#")

mixed_file_text = """
@prefix ex: <http://example.org/ns#> .
@prefix rdf: <http://www.w3.org/1999/02/22-rdf-syntax-ns#> .
@prefix rdfs: <http://www.w3.org/2000/01/rdf-schema#> .
@prefix sh: <http://www.w3.org/ns/shacl#> .
@prefix xsd:  <http://www.w3.org/2001/XMLSchema#> .

ex:myProperty-datatype
	a sh:PropertyShape ;
	rdfs:comment "Violations of sh:datatype are produced as warnings"@en ;
	sh:datatype xsd:string ;
	sh:path ex:myProperty ;
	sh:severity sh:Warning ;
	.

ex:myProperty-maxLength
	a sh:PropertyShape ;
	rdfs:comment "The default severity here is sh:Info"@en ;
	sh:maxLength 10 ;
	sh:message "Too many characters"@en ;
	sh:path ex:myProperty ;
	sh:severity sh:Info ;
	.

ex:MyShape
	a sh:NodeShape ;
	rdfs:comment "This example was adapted from the example shapes and validation graph in the Severity section of the SHACL specification."@en ;
	rdfs:seeAlso "https://www.w3.org/TR/shacl/#severity" ;
	sh:property
		ex:myProperty-datatype ,
		ex:myProperty-maxLength
		;
	sh:targetNode ex:MyInstance ;
	.

ex:MyInstance
	rdfs:comment "This instance triggers one Info-level and one Warning-level violation."@en ;
	ex:myProperty "http://toomanycharacters"^^xsd:anyURI ;
	.
"""


def _test_126_template(expected_conformance: bool, allow_level: Optional[URIRef] = None) -> None:
    validate_kwargs: Dict[Any] = {"data_graph_format": "turtle", "shacl_graph_format": "turtle", "debug": True}
    if allow_level is None:
        pass
    elif allow_level == SH.Info:
        validate_kwargs["allow_infos"] = True
    elif allow_level == SH.Warning:
        validate_kwargs["allow_warnings"] = True
    else:
        raise NotImplementedError("allow_level=%r" % allow_level)

    res1 = validate(mixed_file_text, **validate_kwargs)
    conforms, graph, string = res1
    assert conforms == expected_conformance

    datatype_shape_reported = False
    maxlength_shape_reported = False

    # Confirm the report graph emits ValidationResults pertaining to each expected triggering property shape.
    for triple in graph.triples((None, SH.sourceShape, NS_EX["myProperty-datatype"])):
        datatype_shape_reported = True
    for triple in graph.triples((None, SH.sourceShape, NS_EX["myProperty-maxLength"])):
        maxlength_shape_reported = True

    assert datatype_shape_reported
    assert maxlength_shape_reported


def test_126_1() -> None:
    """
    With no severities allowed, expect non-conformance.
    """
    _test_126_template(False)


def test_126_2() -> None:
    """
    With Infos allowed, expect non-conformance.
    """
    _test_126_template(False, SH.Info)


def test_126_3() -> None:
    """
    With Warnings allowed, expect conformance.
    """
    _test_126_template(True, SH.Warning)



# -*- coding: utf-8 -*-
"""
https://github.com/RDFLib/pySHACL/issues/148
"""

import rdflib

from pyshacl import validate


shacl_file = """\
@prefix ex: <urn:ex#> .
@prefix sh: <http://www.w3.org/ns/shacl#> .
@prefix owl: <http://www.w3.org/2002/07/owl#> .
@prefix rdf: <http://www.w3.org/1999/02/22-rdf-syntax-ns#> .
@prefix rdfs: <http://www.w3.org/2000/01/rdf-schema#> .

ex:Class a owl:Class .
ex:FailedRule a sh:NodeShape ;
    sh:targetClass ex:Class ;
    sh:rule [
        a sh:TripleRule ;
        sh:object ex:Inferred ;
        sh:predicate ex:hasProperty ;
        sh:subject sh:this ;
    ] .
"""
ont_file = """\
@prefix ex: <urn:ex#> .
@prefix owl: <http://www.w3.org/2002/07/owl#> .
@prefix rdf: <http://www.w3.org/1999/02/22-rdf-syntax-ns#> .
@prefix rdfs: <http://www.w3.org/2000/01/rdf-schema#> .

ex:Class a owl:Class .
ex:SubClass a owl:Class ;
    rdfs:subClassOf ex:Class .
ex:SubSubClass a owl:Class ;
    rdfs:subClassOf ex:SubClass .

"""

data_file = """\
@prefix ex: <urn:ex#> .
@prefix rdf: <http://www.w3.org/1999/02/22-rdf-syntax-ns#> .

ex:A a ex:SubSubClass .
"""


def test_148() -> None:
    data_g = rdflib.Graph()
    data_g.parse(data=data_file, format="turtle").parse(data=ont_file)
    shapes = rdflib.Graph()
    shapes.parse(data=shacl_file, format="turtle")
    res = validate(data_g, shacl_graph=shapes, advanced=True)
    assert (
        rdflib.URIRef("urn:ex#A"),
        rdflib.URIRef("urn:ex#hasProperty"),
        rdflib.URIRef("urn:ex#Inferred"),
    ) not in data_g

    conforms, graph, string = res
    assert conforms


if __name__ == "__main__":
    test_148()



# -*- coding: utf-8 -*-
"""
https://github.com/RDFLib/pySHACL/issues/141
"""

import rdflib

from pyshacl import validate


shacl_file = r"""@prefix rdf: <http://www.w3.org/1999/02/22-rdf-syntax-ns#> .
@prefix ex: <http://example.org/ontology/> .
@prefix owl: <http://www.w3.org/2002/07/owl#> .
@prefix rdfs: <http://www.w3.org/2000/01/rdf-schema#> .
@prefix sh: <http://www.w3.org/ns/shacl#> .

ex:Dog
	a owl:Class , sh:NodeShape ;
	sh:property [
		a sh:PropertyShape ;
		sh:minCount 1 ;
		sh:nodeKind sh:BlankNodeOrIRI ;
		sh:path ex:hasIdentifier
	] ;
	sh:targetClass ex:Dog ;
	.

ex:Cat
	a owl:Class , sh:NodeShape	;
	sh:property [
		a sh:PropertyShape ;
		sh:minCount 1 ;
		sh:nodeKind sh:BlankNodeOrIRI ;
		sh:path ex:hasCharacteristic
	] ;
	sh:targetClass ex:Cat ;
	.

ex:hasIdentifier a owl:ObjectProperty ;
	.

ex:hasCharacteristic a owl:ObjectProperty ;
	.

ex:hasName a owl:ObjectProperty ;
	rdfs:subPropertyOf ex:hasIdentifier , ex:hasCharacteristic ;
	.

ex:hasLabel a owl:ObjectProperty ;
	rdfs:subPropertyOf ex:hasIdentifier ;
	.
"""


data_file = """@prefix rdf: <http://www.w3.org/1999/02/22-rdf-syntax-ns#> .
@prefix owl: <http://www.w3.org/2002/07/owl#> .
@prefix ex: <http://example.org/ontology/> .
@prefix kb: <http://example.org/kb/> .
@prefix rdfs: <http://www.w3.org/2000/01/rdf-schema#> .


ex:hasIdentifier a owl:ObjectProperty ;
	.

ex:hasCharacteristic a owl:ObjectProperty ;
	.

ex:hasName a owl:ObjectProperty ;
	rdfs:subPropertyOf ex:hasIdentifier , ex:hasCharacteristic ;
	.

ex:hasLabel a owl:ObjectProperty ;
	rdfs:subPropertyOf ex:hasIdentifier ;
	.

kb:dog_with_identifier
	a ex:Dog ;
	rdfs:comment "PASS" ;
	ex:hasIdentifier kb:dogidentifier1 ;
	.

kb:dog_with_name
	a ex:Dog ;
	rdfs:comment "PASS" ;
	ex:hasName [ rdf:value "Rover" ] ;
	.

kb:dog_with_characteristic
	a ex:Dog ;
	rdfs:comment "XFAIL" ;
	ex:hasCharacteristic [ rdf:value "Brown" ] ;
	.

kb:cat_with_identifier
	a ex:Cat ;
	rdfs:comment "XFAIL" ;
	ex:hasIdentifier kb:catidentifier1 ;
	.

kb:cat_with_name
	a ex:Cat ;
	rdfs:comment "PASS" ;
	ex:hasName [ rdf:value "Mittens" ] ;
	.

kb:cat_with_characteristic
	a ex:Cat ;
	rdfs:comment "PASS" ;
	ex:hasCharacteristic [ rdf:value "Green" ] ;
	.

kb:dog_with_label
	a ex:Dog ;
	rdfs:comment "PASS" ;
	ex:hasLabel [ rdf:value "Dopey" ] ;
	.
"""


def test_me() -> None:
    my_json = """\
    {
  "@context": {
    "@vocab": "http://schema.org/",
    "xsd": "http://www.w3.org/2001/XMLSchema#"
  },
  "@type": "MonetaryAmount",
  "value": {
    "@type": "xsd:float",
    "@value": "100000"
  }
}
    """
    shapes_json = """\
    {
  "@context": {
    "owl": "http://www.w3.org/2002/07/owl#",
    "rdf": "http://www.w3.org/1999/02/22-rdf-syntax-ns#",
    "rdfs": "http://www.w3.org/2000/01/rdf-schema#",
    "xsd": "http://www.w3.org/2001/XMLSchema#",
    "skos": "http://www.w3.org/2004/02/skos/core#",
    "prov": "http://www.w3.org/ns/prov#",
    "dcat": "http://www.w3.org/ns/dcat#",
    "sh": "http://www.w3.org/ns/shacl#",
    "shsh": "http://www.w3.org/ns/shacl-shacl#",
    "dcterms": "http://purl.org/dc/terms/",
    "schema": "http://schema.org/",
    "rescs": "http://rescs.org/"
  },
  "@graph": [
    {
      "@id": "rescs:dash/monetaryamount/MonetaryAmountShape",
      "@type": "sh:NodeShape",
      "rdfs:comment": {
        "@type": "xsd:string",
        "@value": "A monetary value or range. This type can be used to describe an amount of money such as $50 USD, or a range as in describing a bank account being suitable for a balance between £1,000 and £1,000,000 GBP, or the value of a salary, etc. It is recommended to use [[PriceSpecification]] Types to describe the price of an Offer, Invoice, etc."
      },
      "rdfs:label": {
        "@type": "xsd:string",
        "@value": "Monetary amount"
      },
      "sh:property": {
        "sh:datatype": {
          "@id": "xsd:float"
        },
        "sh:description": "The value of the quantitative value or property value node.\\\\n\\\\n* For [[QuantitativeValue]] and [[MonetaryAmount]], the recommended type for values is 'Number'.\\\\n* For [[PropertyValue]], it can be 'Text;', 'Number', 'Boolean', or 'StructuredValue'.\\\\n* Use values from 0123456789 (Unicode 'DIGIT ZERO' (U+0030) to 'DIGIT NINE' (U+0039)) rather than superficially similiar Unicode symbols.\\\\n* Use '.' (Unicode 'FULL STOP' (U+002E)) rather than ',' to indicate a decimal point. Avoid using these symbols as a readability separator.",
        "sh:maxCount": {
          "@type": "xsd:integer",
          "@value": 1
        },
        "sh:minCount": {
          "@type": "xsd:integer",
          "@value": 1
        },
        "sh:minExclusive": 0,
        "sh:name": "value",
        "sh:path": {
          "@id": "schema:value"
        }
      },
      "sh:targetClass": {
        "@id": "schema:MonetaryAmount"
      }
    }
  ]
}
    """
    g = rdflib.Graph()
    g.parse(data=shapes_json, format="json-ld")
    d = rdflib.Graph()
    d.parse(data=my_json, format="json-ld")
    validate(d, shacl_graph=g, debug=True)


def test_141() -> None:
    data = rdflib.Graph()
    data.parse(data=data_file, format="turtle")
    shapes = rdflib.Graph()
    shapes.parse(data=shacl_file, format="turtle")
    res = validate(
        data,
        shacl_graph=shapes,
        data_graph_format='turtle',
        shacl_graph_format='turtle',
        debug=True,
        inference="rdfs",
    )

    conforms, graph, string = res


if __name__ == "__main__":
    test_141()



#
#
import typing
from typing import Any, Dict, List, Tuple, Union

from rdflib import Literal, URIRef

from pyshacl.constraints import ConstraintComponent
from pyshacl.constraints.constraint_component import CustomConstraintComponent
from pyshacl.consts import SH, SH_ConstraintComponent, SH_message
from pyshacl.errors import ConstraintLoadError, ReportableRuntimeError, ValidationFailure
from pyshacl.pytypes import GraphLike, SHACLExecutor

from .js_executable import JSExecutable

if typing.TYPE_CHECKING:
    from pyshacl.shape import Shape
    from pyshacl.shapes_graph import ShapesGraph


SH_JSConstraint = SH.JSConstraint
SH_JSConstraintComponent = SH.JSConstraintComponent


class BoundShapeJSValidatorComponent(ConstraintComponent):
    invalid_parameter_names = {'this', 'shapesGraph', 'currentShape', 'path', 'PATH', 'value'}

    shacl_constraint_component = SH_ConstraintComponent

    def __init__(self, constraint, shape: 'Shape', validator):
        """
        Create a new custom constraint, by applying a ConstraintComponent and a Validator to a Shape
        :param constraint: The source ConstraintComponent, this is needed to bind the parameters in the query_helper
        :type constraint: SPARQLConstraintComponent
        :param shape:
        :type shape: Shape
        :param validator:
        :type validator: AskConstraintValidator | SelectConstraintValidator
        """
        super(BoundShapeJSValidatorComponent, self).__init__(shape)
        self.constraint = constraint
        self.validator = validator
        self.param_bind_map: Dict[str, Any] = {}
        self.messages: List[Any] = []
        self.bind_params()

    def bind_params(self):
        bind_map = {}
        shape = self.shape
        for p in self.constraint.parameters:
            name = p.localname
            if name in self.invalid_parameter_names:
                # TODO:coverage: No test for this case
                raise ReportableRuntimeError("Parameter name {} cannot be used.".format(name))
            shape_params = set(shape.objects(p.path()))
            if len(shape_params) < 1:
                if not p.optional:
                    # TODO:coverage: No test for this case
                    raise ReportableRuntimeError("Shape does not have mandatory parameter {}.".format(str(p.path())))
                continue
            # TODO: Can shapes have more than one value for the predicate?
            # Just use one for now.
            # TODO: Check for sh:class and sh:nodeKind on the found param value
            bind_map[name] = next(iter(shape_params))
        self.param_bind_map = bind_map

    @classmethod
    def constraint_parameters(cls) -> List[URIRef]:
        # TODO:coverage: this is never used for this constraint?
        return []

    @classmethod
    def constraint_name(cls) -> str:
        return "ConstraintComponent"

    def make_generic_messages(self, datagraph: GraphLike, focus_node, value_node) -> List[Literal]:
        return [Literal("Parameterised Javascript Function generated constraint validation reports.")]

    def evaluate(
        self, executor: SHACLExecutor, data_graph: GraphLike, focus_value_nodes: Dict, _evaluation_path: List
    ):
        """
        :type executor: SHACLExecutor
        :type data_graph: rdflib.Graph
        :type focus_value_nodes: dict
        :type _evaluation_path: list
        """
        reports = []
        non_conformant = False
        extra_messages = self.messages or []
        rept_kwargs = {
            'constraint_component': self.constraint.node,
            'extra_messages': extra_messages,
        }
        for f, value_nodes in focus_value_nodes.items():
            try:
                p = self.shape.path()
                results = self.validator.validate(f, value_nodes, p, data_graph, self.param_bind_map)
            except ValidationFailure as e:
                raise e
            for v, result in results:
                if result is True:
                    continue
                args_map = self.param_bind_map.copy()
                args_map.update({"this": f, "value": v})
                if self.shape.is_property_shape:
                    args_map['path'] = self.shape.path()
                bound_messages = self.validator.make_messages(args_map)
                failed = False
                if isinstance(result, list):
                    pass
                else:
                    result = [result]
                for res in result:
                    if isinstance(res, Literal):
                        res = res.value
                    if isinstance(res, bool):
                        if res:
                            continue
                        else:
                            failed = True
                            new_kwargs = rept_kwargs.copy()
                            new_kwargs['extra_messages'].extend(bound_messages)
                            reports.append(self.make_v_result(data_graph, f, value_node=v, **new_kwargs))
                    elif isinstance(res, str):
                        failed = True
                        m = Literal(res)
                        new_kwargs = rept_kwargs.copy()
                        new_kwargs['extra_messages'].append(m)
                        new_kwargs['extra_messages'].extend(bound_messages)
                        reports.append(self.make_v_result(data_graph, f, value_node=v, **new_kwargs))
                    elif isinstance(res, dict):
                        failed = True
                        args_map2 = args_map.copy()
                        val = res.get('value', None)
                        if val is None:
                            val = v
                        args_map2['value'] = val
                        path = res.get('path', None) if not self.shape.is_property_shape else None
                        if path is not None:
                            args_map2['value'] = path
                        msgs = self.validator.make_messages(args_map2)
                        message = res.get('message', None)
                        if message is not None:
                            msgs.append(Literal(message))
                        new_kwargs = rept_kwargs.copy()
                        new_kwargs['extra_messages'].extend(msgs)
                        reports.append(
                            self.make_v_result(data_graph, f, value_node=val, result_path=path, **new_kwargs)
                        )
                if failed:
                    non_conformant = True
        return (not non_conformant), reports


class JSConstraintComponent(CustomConstraintComponent):
    """
    SPARQL-based constraints provide a lot of flexibility but may be hard to understand for some people or lead to repetition. This section introduces SPARQL-based constraint components as a way to abstract the complexity of SPARQL and to declare high-level reusable components similar to the Core constraint components. Such constraint components can be declared using the SHACL RDF vocabulary and thus shared and reused.
    Link:
    https://www.w3.org/TR/shacl-js/#js-components
    """

    __slots__: Tuple = tuple()

    def __new__(cls, shacl_graph, node, parameters, validators, node_validators, property_validators):
        return super(JSConstraintComponent, cls).__new__(
            cls, shacl_graph, node, parameters, validators, node_validators, property_validators
        )

    def make_validator_for_shape(self, shape: 'Shape'):
        """
        :param shape:
        :type shape: Shape
        :return:
        """
        val_count = len(self.validators)
        node_val_count = len(self.node_validators)
        prop_val_count = len(self.property_validators)
        is_property_val = False
        if shape.is_property_shape and prop_val_count > 0:
            validator_node = next(iter(self.property_validators))
            is_property_val = True
        elif (not shape.is_property_shape) and node_val_count > 0:
            validator_node = next(iter(self.node_validators))
        elif val_count > 0:
            validator_node = next(iter(self.validators))
        else:
            raise ConstraintLoadError(
                "Cannot select a validator to use, according to the rules.",
                "https://www.w3.org/TR/shacl/#constraint-components-validators",
            )
        if is_property_val:
            validator: Union[JSConstraintComponentValidator, JSConstraintComponentPathValidator] = (
                JSConstraintComponentPathValidator(self.sg, validator_node)
            )
        else:
            validator = JSConstraintComponentValidator(self.sg, validator_node)
        applied_validator = validator.apply_to_shape_via_constraint(self, shape)
        return applied_validator


class JSConstraintComponentValidator(JSExecutable):
    __slots__ = ("messages", "initialised")

    validator_cache: Dict[Tuple[int, str], 'JSConstraintComponentValidator'] = {}

    def __new__(cls, shacl_graph: 'ShapesGraph', node, *args, **kwargs):
        cache_key = (id(shacl_graph.graph), str(node))
        found_in_cache = cls.validator_cache.get(cache_key, False)
        if found_in_cache:
            return found_in_cache
        self = super(JSConstraintComponentValidator, cls).__new__(cls, shacl_graph, node)
        cls.validator_cache[cache_key] = self
        return self

    def __init__(self, shacl_graph: 'ShapesGraph', node, *args, **kwargs):
        initialised = getattr(self, 'initialised', False)
        if initialised:
            return
        super(JSConstraintComponentValidator, self).__init__(shacl_graph, node)
        sg = shacl_graph.graph
        message_nodes = set(sg.objects(node, SH_message))
        for m in message_nodes:
            if not (isinstance(m, Literal) and isinstance(m.value, str)):
                # TODO:coverage: No test for when SPARQL-based constraint is RDF Literal is is not of type string
                raise ConstraintLoadError(
                    "Validator sh:message must be an RDF Literal of type xsd:string.",
                    "https://www.w3.org/TR/shacl/#ConstraintComponent",
                )
        self.messages = message_nodes
        self.initialised = True

    def make_messages(self, args_map=None):
        if args_map is None:
            return self.messages
        ret_msgs = []
        for m in self.messages:
            this_m = m.value[:]
            for a, v in args_map.items():
                replace_me = "{$" + str(a) + "}"
                if isinstance(v, Literal):
                    v = v.value
                this_m = this_m.replace(replace_me, str(v))
            ret_msgs.append(Literal(this_m))
        return ret_msgs

    def validate(self, f, value_nodes, path, data_graph, param_bind_vals, new_bind_vals=None):
        """

        :param f:
        :param value_nodes:
        :param path:
        :param data_graph:
        :type data_graph: rdflib.Graph
        :param new_bind_vals:
        :return:
        """
        new_bind_vals = new_bind_vals or {}
        bind_vals = param_bind_vals.copy()
        bind_vals.update(new_bind_vals)
        results = []
        for v in value_nodes:
            args_map = bind_vals.copy()
            args_map.update({'this': f, 'value': v})
            try:
                result_dict = self.execute(data_graph, args_map)
                results.append((v, result_dict['_result']))
            except Exception as e:
                print(e)
                raise
        return results

    def apply_to_shape_via_constraint(self, constraint, shape, **kwargs) -> BoundShapeJSValidatorComponent:
        """
        Create a new Custom Constraint (BoundShapeValidatorComponent)
        :param constraint:
        :type constraint: JSConstraintComponent
        :param shape:
        :type shape: pyshacl.shape.Shape
        :param kwargs:
        :return:
        :rtype: BoundShapeJSValidatorComponent
        """
        return BoundShapeJSValidatorComponent(constraint, shape, self)


class JSConstraintComponentPathValidator(JSConstraintComponentValidator):
    path_validator_cache: Dict[Tuple[int, str], 'JSConstraintComponentPathValidator'] = {}

    def __new__(cls, shacl_graph: 'ShapesGraph', node, *args, **kwargs):
        cache_key = (id(shacl_graph.graph), str(node))
        found_in_cache = cls.path_validator_cache.get(cache_key, False)
        if found_in_cache:
            return found_in_cache
        self = super(JSConstraintComponentPathValidator, cls).__new__(cls, shacl_graph, node)
        cls.path_validator_cache[cache_key] = self
        return self

    def validate(self, f, value_nodes, path, data_graph, param_bind_vals, new_bind_vals=None):
        """

        :param f:
        :param value_nodes:
        :param path:
        :param data_graph:
        :type data_graph: rdflib.Graph
        :param new_bind_vals:
        :return:
        """
        new_bind_vals = new_bind_vals or {}
        args_map = param_bind_vals.copy()
        args_map.update(new_bind_vals)
        args_map.update({'this': f, 'path': path})
        results = []
        try:
            result_dict = self.execute(data_graph, args_map)
            results.append((f, result_dict['_result']))
        except Exception as e:
            print(e)
            raise
        return results



import pickle

from rdflib import Graph, URIRef
from rdflib.plugins.stores.memory import Memory

from pyshacl.monkey import apply_patches

apply_patches()
identifier = URIRef("http://datashapes.org/schema")
store = Memory(identifier=identifier)
with open("./schema.ttl", "rb") as f:
    g = Graph(store=store, identifier=identifier, bind_namespaces='core').parse(file=f)
with open("./schema.pickle", "wb") as f:
    pickle.dump((store, identifier), f, protocol=5)

identifier = URIRef("http://www.w3.org/ns/shacl#")
store = Memory(identifier=identifier)
with open("./shacl.ttl", "rb") as f:
    g = Graph(store=store, identifier=identifier, bind_namespaces='core').parse(file=f)
with open("./shacl.pickle", "wb") as f:
    pickle.dump((store, identifier), f, protocol=5)

identifier = URIRef("http://datashapes.org/dash")
store = Memory(identifier=identifier)
with open("./dash.ttl", "rb") as f:
    g = Graph(store=store, identifier=identifier, bind_namespaces='core').parse(file=f)
with open("./dash.pickle", "wb") as f:
    pickle.dump((store, identifier), f, protocol=5)

identifier = URIRef("http://www.w3.org/ns/shacl-shacl#")
store = Memory(identifier=identifier)
with open("./shacl-shacl.ttl", "rb") as f:
    g = Graph(store=store, identifier=identifier, bind_namespaces='core').parse(file=f)
with open("./shacl-shacl.pickle", "wb") as f:
    pickle.dump((store, identifier), f, protocol=5)



# -*- coding: utf-8 -*-
#
"""
https://github.com/RDFLib/pySHACL/issues/240
"""
import sys

import rdflib

import pyshacl

shapes_data = '''\
@prefix dash: <http://datashapes.org/dash#> .
@prefix rdf: <http://www.w3.org/1999/02/22-rdf-syntax-ns#> .
@prefix rdfs: <http://www.w3.org/2000/01/rdf-schema#> .
@prefix schema: <http://schema.org/> .
@prefix sh: <http://www.w3.org/ns/shacl#> .
@prefix xsd: <http://www.w3.org/2001/XMLSchema#> .
@prefix ex: <http://example.org/ns#> .

schema:PersonShape
    a sh:NodeShape ;
    sh:targetClass schema:Family;
sh:property schema:family-familyname.

schema:family-familyname
        rdf:type     sh:PropertyShape ;
        sh:path      schema:CorrectfamilyName ;
        sh:name      "familyname" ;
        sh:sparql    schema:onefamilyname ;
        sh:nodeKind  sh:Literal .

schema:onefamilyname
        rdfs:label  "onefamilyname " ;
        rdf:type    sh:SPARQLConstraint ;
        sh:message  "{?this} {?person}" ;
        sh:select   """prefix dash: <http://datashapes.org/dash#>
        prefix rdf: <http://www.w3.org/1999/02/22-rdf-syntax-ns#>
        prefix rdfs: <http://www.w3.org/2000/01/rdf-schema#>
        prefix schema: <http://schema.org/>
        prefix sh: <http://www.w3.org/ns/shacl#>
        prefix xsd: <http://www.w3.org/2001/XMLSchema#>
        prefix ex: <http://example.org/ns#>

        SELECT ?this ?person
        WHERE {
          ?this a schema:Family.
          ?person schema:family ?this.
        }""" .

'''

data_g_text = '''\
@prefix ex: <http://example.org/ns#> .
@prefix rdf: <http://www.w3.org/1999/02/22-rdf-syntax-ns#> .
@prefix rdfs: <http://www.w3.org/2000/01/rdf-schema#> .
@prefix schema: <http://schema.org/> .
@prefix xsd: <http://www.w3.org/2001/XMLSchema#> .

ex:Bob
    a schema:Person ;
    schema:givenName "Robert" ;
    schema:familyName "Junior" ;
	schema:family ex:JuniorFamily .

ex:Molly
    a schema:Person ;
    schema:givenName "Molly" ;
    schema:familyName "Junior2" ;
	schema:family ex:JuniorFamily .

ex:John
    a schema:Person ;
    schema:givenName "John" ;
    schema:familyName "Junior3" ;
	schema:family ex:JuniorFamily .


ex:JuniorFamily
	a schema:Family;
	schema:CorrectfamilyName "Junior" .
'''


def test_240():
    shape_g = rdflib.Graph().parse(data=shapes_data, format='turtle')
    data_g = rdflib.Graph().parse(data=data_g_text, format="turtle")
    conforms, results_graph, results_text = pyshacl.validate(
        data_g,
        shacl_graph=shape_g,
        debug=True,
        inference='none',
        advanced=True,
        meta_shacl=False,
    )
    assert not conforms
    assert "Results (3)" in results_text


if __name__ == "__main__":
    sys.exit(test_240())
